# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generic helpers for ord_schema, including common message types."""
from typing import Union

from google.protobuf import descriptor
import google.protobuf.message

from ord_schema.proto import reaction_pb2

# Commonly used types.
FieldDescriptor = descriptor.FieldDescriptor
Message = google.protobuf.message.Message
ScalarType = Union[str, bytes, float, int, bool]

# Messages with 'type' and 'details' fields.
TypeDetailsMessage = Union[
    reaction_pb2.Analysis,
    reaction_pb2.CompoundIdentifier,
    reaction_pb2.CompoundPreparation,
    reaction_pb2.ElectrochemistryConditions,
    reaction_pb2.ElectrochemistryConditions.ElectrochemistryCell,
    reaction_pb2.FlowConditions,
    reaction_pb2.FlowConditions.Tubing,
    reaction_pb2.IlluminationConditions,
    reaction_pb2.PressureConditions.Atmosphere,
    reaction_pb2.PressureConditions.Measurement,
    reaction_pb2.PressureConditions.PressureControl,
    reaction_pb2.ProductCompound.Texture,
    reaction_pb2.ProductMeasurement,
    reaction_pb2.ProductMeasurement.MassSpecMeasurementDetails,
    reaction_pb2.ProductMeasurement.Selectivity,
    reaction_pb2.ReactionIdentifier,
    reaction_pb2.ReactionInput.AdditionDevice,
    reaction_pb2.ReactionSetup.ReactionEnvironment,
    reaction_pb2.ReactionWorkup,
    reaction_pb2.StirringConditions,
    reaction_pb2.TemperatureConditions.Measurement,
    reaction_pb2.TemperatureConditions.TemperatureControl,
    reaction_pb2.Vessel,
    reaction_pb2.VesselAttachment,
    reaction_pb2.VesselMaterial,
    reaction_pb2.VesselPreparation,
]

# Messages with 'units' fields.
UnitMessage = Union[
    reaction_pb2.Current,
    reaction_pb2.FlowRate,
    reaction_pb2.Length,
    reaction_pb2.Mass,
    reaction_pb2.Moles,
    reaction_pb2.Pressure,
    reaction_pb2.Temperature,
    reaction_pb2.Time,
    reaction_pb2.Voltage,
    reaction_pb2.Volume,
    reaction_pb2.Wavelength,
]



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base ORM objects."""
from inspect import getmro
from inflection import underscore
from sqlalchemy import Column, Integer, ForeignKey, String
from sqlalchemy.orm import declarative_base, declarative_mixin, declared_attr


class Base:
    """See https://docs.sqlalchemy.org/en/14/orm/declarative_mixins.html#augmenting-the-base.

    * The table name is a snake_case rendering of the class name.
    * Every table has an `id` column used as the primary key. Note that _Child tables do not have this column.
    """

    id = Column(Integer, primary_key=True)

    @declared_attr
    def __tablename__(cls):  # pylint: disable=no-self-argument
        name = cls.__name__  # pylint: disable=no-member
        if name.startswith("_"):
            name = name[1:]  # Remove leading underscore.
        return underscore(name)  # pylint: disable=no-member


Base = declarative_base(cls=Base)


@declarative_mixin
class Parent:
    """Mixin class for parent classes.

    * Polymorphic types are identified in the `_type` column.
    """

    _type = Column(String(255), nullable=False)

    @declared_attr
    def __mapper_args__(cls):  # pylint: disable=no-self-argument
        name = cls.__name__  # pylint: disable=no-member
        if name.startswith("_"):
            name = name[1:]  # Remove leading underscore.
        return {
            "polymorphic_identity": underscore(name),  # pylint: disable=no-member
            "polymorphic_on": cls._type,
        }


@declarative_mixin
class Child:
    """Mixin class for child classes.

    * The parent ID is stored in the `parent_id` column (also used as the primary key).
    """

    @declared_attr
    def parent_id(cls):  # pylint: disable=no-self-argument
        """Creates a `parent_id` column referring to the parent table."""
        parent_table = None
        for parent_class in getmro(cls)[1:]:
            if issubclass(parent_class, Base):
                parent_table = parent_class.__tablename__
                break
        assert parent_table is not None, (cls, getmro(cls))
        if parent_table in ["structure"]:
            foreign_key = ForeignKey(f"rdkit.{parent_table}.id")
        else:
            foreign_key = ForeignKey(f"{parent_table}.id")
        return Column(f"{parent_table}_id", Integer, foreign_key, primary_key=True)

    @declared_attr
    def __mapper_args__(cls):  # pylint: disable=no-self-argument
        return {"polymorphic_identity": underscore(cls.__name__)}  # pylint: disable=no-member



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pytest fixtures."""
import os
from typing import Iterator

import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from testing.postgresql import Postgresql

from ord_schema.message_helpers import load_message
from ord_schema.orm.database import add_datasets, add_rdkit, prepare_database
from ord_schema.proto.dataset_pb2 import Dataset


@pytest.fixture
def test_session() -> Iterator[Session]:
    dataset = load_message(os.path.join(os.path.dirname(__file__), "testdata", "ord-nielsen-example.pbtxt"), Dataset)
    with Postgresql() as postgres:
        engine = create_engine(postgres.url(), echo=False, future=True)
        rdkit_cartridge = prepare_database(engine)
        with Session(engine) as session:
            add_datasets([dataset], session)
            session.flush()
            if rdkit_cartridge:
                add_rdkit(session)
            session.commit()
        with Session(engine) as session:
            yield session



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.proto.reaction_pb2."""

import pytest

from ord_schema.proto import reaction_pb2


def test_simple():
    reaction = reaction_pb2.Reaction()
    reaction.identifiers.add(value="C(C)Cl.Br>>C(C)Br.Cl", type="REACTION_SMILES")
    assert reaction.IsInitialized()
    assert len(reaction.identifiers) == 1
    assert not reaction.HasField("setup")
    with pytest.raises(ValueError, match="not_a_field"):
        reaction.HasField("not_a_field")



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset processing script for database submissions.

This script is meant to be a one-stop shop for preparing submissions to the
Open Reaction Database.

By default, the script only validates the input Dataset messages. Validation
may introduce changes to the Reaction messages, such as the addition of SMILES
for compounds identified only by NAME. Users should frequently run these checks
as they are preparing a dataset for submission.

With the optional --update flag, the script also performs database-specific
updates (such as adding record IDs). These operations are meant to be run as
part of the submission process and not as part of the pre-submission validation
cycle.

Usage:
    process_dataset.py (--input_pattern=<str> | --input_file=<str>) [options]

Options:
    --input_pattern=<str>       Pattern matching input Dataset protos
    --input_file=<str>          File containing Dataset proto filenames
    --root=<str>                Root of the repository [default: ]
    --output_format=<str>       Dataset output format [default: .pb.gz]
    --write_errors              If True, errors will be written to *.error
    --no-validate               If set, reactions will not be validated
    --update                    If True, update Reaction protos
    --cleanup                   If True, use git to clean up
    --max_size=<float>          Maximum size (in MB) for any Reaction message [default: 10.0]
    --base=<str>                Git branch to diff against
    --issue=<str>               GitHub pull request number; if provided, a comment will be added
    --token=<str>               GitHub authentication token
"""
import dataclasses
import glob
import gzip
import os
import subprocess
import sys
from collections.abc import Iterable, Mapping
from typing import Optional

import docopt
import github
from rdkit import RDLogger

from ord_schema.logging import get_logger
from ord_schema import message_helpers
from ord_schema import updates
from ord_schema import validations
from ord_schema.proto import dataset_pb2

logger = get_logger(__name__)


# pylint: disable=too-many-branches,too-many-locals


@dataclasses.dataclass(eq=True, frozen=True, order=True)
class FileStatus:
    """A filename and its status in Git."""

    filename: str
    status: str
    original_filename: str

    def __post_init__(self):
        if self.status[0] not in ["A", "D", "M", "R"]:
            raise ValueError(f"unsupported file status: {self.status}")


def _get_inputs(kwargs) -> list[FileStatus]:
    """Gets a list of Dataset proto filenames to process.

    Returns:
        List of FileStatus objects.

    Raises:
        ValueError: If a git-diff status is not one of {'A', 'D', 'M', 'R'}.
    """
    if kwargs["--input_pattern"]:
        # Setting recursive=True allows recursive matching with '**'.
        filenames = glob.glob(kwargs["--input_pattern"], recursive=True)
        return [FileStatus(filename, "A", "") for filename in filenames]
    if kwargs["--input_file"]:
        inputs = []
        with open(kwargs["--input_file"]) as f:
            for line in f:
                fields = line.strip().split("\t")
                if len(fields) == 3:
                    status, original_filename, filename = fields
                    if not status.startswith("R"):
                        raise ValueError(f"malformed status line: {line.strip()}")
                else:
                    status, filename = fields
                    if not status.startswith(("A", "D", "M")):
                        raise ValueError(f"unsupported git-diff statue: {status}")
                    original_filename = ""
                inputs.append(FileStatus(filename, status, original_filename))
        return inputs
    raise ValueError("one of --input_pattern or --input_file is required")


def cleanup(filename: str, output_filename: str):
    """Removes and/or renames the input Dataset files.

    Args:
        filename: Original dataset filename.
        output_filename: Updated dataset filename.
    """
    if filename == output_filename:
        logger.info("editing an existing dataset; no cleanup needed")
        return  # Reuse the existing dataset ID.
    args = ["git", "mv", filename, output_filename]
    logger.info("Running command: %s", " ".join(args))
    subprocess.run(args, check=True)


def _get_reaction_ids(dataset: dataset_pb2.Dataset) -> set[str]:
    """Returns a set containing the reaction IDs in a Dataset."""
    reaction_ids = set()
    for reaction in dataset.reactions:
        if reaction.reaction_id:
            reaction_ids.add(reaction.reaction_id)
    return reaction_ids


def _load_base_dataset(file_status: FileStatus, base: str) -> dataset_pb2.Dataset:
    """Loads a Dataset message from another branch."""
    if file_status.status.startswith("A"):
        return None  # Dataset only exists in the submission.
    # NOTE(kearnes): Use --no-pager to avoid a non-zero exit code.
    args = ["git", "--no-pager", "show"]
    if file_status.status.startswith("R"):
        args.append(f"{base}:{file_status.original_filename}")
    else:
        args.append(f"{base}:{file_status.filename}")
    logger.info("Running command: %s", " ".join(args))
    serialized = subprocess.run(args, capture_output=True, check=True, text=False)
    if serialized.stdout.startswith(b"version"):
        # Convert Git LFS pointers to real data.
        serialized = subprocess.run(
            ["git", "lfs", "smudge"],
            input=serialized.stdout,
            capture_output=True,
            check=True,
            text=False,
        )
    if args[-1].endswith(".gz"):
        value = gzip.decompress(serialized.stdout)
    else:
        value = serialized.stdout
    return dataset_pb2.Dataset.FromString(value)


def get_change_stats(
    datasets: Mapping[str, dataset_pb2.Dataset], inputs: Iterable[FileStatus], base: str
) -> tuple[set[str], set[str], set[str]]:
    """Computes diff statistics for the submission.

    Args:
        datasets: Dict mapping filenames to Dataset messages.
        inputs: List of FileStatus objects.
        base: Git branch to diff against.

    Returns:
        added: Set of added reaction IDs.
        removed: Set of deleted reaction IDs.
        changed: Set of changed reaction IDs.
    """
    old, new = set(), set()
    for file_status in inputs:
        if not file_status.status.startswith("D"):
            new.update(_get_reaction_ids(datasets[file_status.filename]))
        dataset = _load_base_dataset(file_status, base)
        if dataset is not None:
            old.update(_get_reaction_ids(dataset))
    return new - old, old - new, new & old


def _run_updates(datasets: Mapping[str, dataset_pb2.Dataset], kwargs) -> None:
    """Updates the submission files.

    Args:
        datasets: Dict mapping filenames to Dataset messages.

    Raises:
        ValueError: if any Reaction is larger than FLAGS.max_size.
    """
    for dataset in datasets.values():
        # Set reaction_ids, resolve names, fix cross-references, etc.
        updates.update_dataset(dataset)
    # Final validation to make sure we didn't break anything.
    options = validations.ValidationOptions(validate_ids=True, require_provenance=True)
    validations.validate_datasets(datasets, kwargs["--write_errors"], options=options)
    for filename, dataset in datasets.items():
        output_filename = os.path.join(
            kwargs["--root"],
            message_helpers.id_filename(f'{dataset.dataset_id}{kwargs["--output_format"]}'),
        )
        os.makedirs(os.path.dirname(output_filename), exist_ok=True)
        if kwargs["--cleanup"]:
            cleanup(filename, output_filename)
        logger.info("writing Dataset to %s", output_filename)
        message_helpers.write_message(dataset, output_filename)


def run(kwargs) -> tuple[Optional[set[str]], Optional[set[str]], Optional[set[str]]]:
    """Main function that returns added/removed reaction ID sets.

    This function should be called directly by tests to get access to the
    return values. If main() returns something other than None it will break
    shell error code logic downstream.

    Returns:
        added: Set of added reaction IDs.
        removed: Set of deleted reaction IDs.
        changed: Set of changed reaction IDs.
    """
    inputs = sorted(_get_inputs(kwargs))
    if not inputs:
        logger.info("nothing to do")
        return set(), set(), set()  # Nothing to do.
    # NOTE(kearnes): Process one dataset at a time to avoid OOM errors.
    change_stats = {}
    for file_status in inputs:
        if file_status.status == "D":
            dataset = None
        else:
            dataset = message_helpers.load_message(file_status.filename, dataset_pb2.Dataset)
            logger.info("%s: %d reactions", file_status.filename, len(dataset.reactions))
        datasets = {file_status.filename: dataset}
        if not kwargs["--no-validate"] and dataset is not None:
            # Note: this does not check if IDs are malformed.
            validations.validate_datasets(datasets, kwargs["--write_errors"])
            # Check reaction sizes.
            for reaction in dataset.reactions:
                reaction_size = sys.getsizeof(reaction.SerializeToString()) / 1e6
                if reaction_size > float(kwargs["--max_size"]):
                    raise ValueError(
                        "Reaction is larger than --max_size " f'({reaction_size} vs {kwargs["--max_size"]}'
                    )
        if kwargs["--base"]:
            added, removed, changed = get_change_stats(datasets, [file_status], base=kwargs["--base"])
            change_stats[file_status.filename] = (added, removed, changed)
            logger.info(
                "Summary: +%d -%d Δ%d reaction IDs",
                len(added),
                len(removed),
                len(changed),
            )
        if kwargs["--update"] and dataset is not None:
            _run_updates(datasets, kwargs)
    if change_stats:
        total_added, total_removed, total_changed = set(), set(), set()
        comment = [
            "Change summary:",
            "| Filename | Added | Removed | Changed |",
            "| -------- | ----- | ------- | ------- |",
        ]
        for filename, (added, removed, changed) in change_stats.items():
            comment.append(f"| {filename} | " f"{len(added)} | {len(removed)} | {len(changed)} |")
            total_added |= added
            total_removed |= removed
            total_changed |= changed
        comment.append(f"| | **{len(total_added)}** | " f"**{len(total_removed)}** | " f"**{len(total_changed)}** |")
        if kwargs["--issue"] and kwargs["--token"]:
            client = github.Github(kwargs["--token"])
            repo = client.get_repo(os.environ["GITHUB_REPOSITORY"])
            issue = repo.get_issue(kwargs["--issue"])
            issue.create_comment("\n".join(comment))
    else:
        total_added, total_removed, total_changed = None, None, None
    return total_added, total_removed, total_changed


def main(kwargs):
    RDLogger.DisableLog("rdApp.*")  # Disable RDKit logging.
    run(kwargs)


if __name__ == "__main__":
    main(docopt.docopt(__doc__))



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.message_helpers."""

import tempfile
import time

import pandas as pd
import pytest
from google.protobuf import json_format
from google.protobuf import text_format
from rdkit import Chem

from ord_schema import message_helpers
from ord_schema.proto import reaction_pb2
from ord_schema.proto import test_pb2

_BENZENE_MOLBLOCK = """241
  -OEChem-07232015262D

 12 12  0     0  0  0  0  0  0999 V2000
    2.8660    1.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.0000    0.5000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    3.7320    0.5000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.0000   -0.5000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    3.7320   -0.5000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.8660   -1.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.8660    1.6200    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    1.4631    0.8100    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    4.2690    0.8100    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    1.4631   -0.8100    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    4.2690   -0.8100    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
    2.8660   -1.6200    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0  0  0  0
  1  3  1  0  0  0  0
  1  7  1  0  0  0  0
  2  4  1  0  0  0  0
  2  8  1  0  0  0  0
  3  5  2  0  0  0  0
  3  9  1  0  0  0  0
  4  6  2  0  0  0  0
  4 10  1  0  0  0  0
  5  6  1  0  0  0  0
  5 11  1  0  0  0  0
  6 12  1  0  0  0  0
M  END"""

# pylint: disable=no-self-use


class TestMessageHelpers:
    @pytest.mark.parametrize(
        "filename,expected",
        (
            ("ord-1234567890", "data/12/ord-1234567890"),
            ("test/ord-foo.pbtxt", "data/fo/ord-foo.pbtxt"),
            ("ord_dataset-f00.pbtxt", "data/f0/ord_dataset-f00.pbtxt"),
            ("ord_data-123456foo7.jpg", "data/12/ord_data-123456foo7.jpg"),
        ),
    )
    def test_id_filename(self, filename, expected):
        assert message_helpers.id_filename(filename) == expected

    @pytest.mark.parametrize(
        "value,identifier_type,expected",
        (
            ("c1ccccc1", "SMILES", "c1ccccc1"),
            ("InChI=1S/C6H6/c1-2-4-6-5-3-1/h1-6H", "INCHI", "c1ccccc1"),
            (_BENZENE_MOLBLOCK, "MOLBLOCK", "c1ccccc1"),
        ),
    )
    def test_smiles_from_compound(self, value, identifier_type, expected):
        compound = reaction_pb2.Compound()
        compound.identifiers.add(value=value, type=identifier_type)
        assert message_helpers.smiles_from_compound(compound) == expected

    @pytest.mark.parametrize(
        "value,identifier_type,expected",
        (
            ("c1ccccc1", "SMILES", "c1ccccc1"),
            ("InChI=1S/C6H6/c1-2-4-6-5-3-1/h1-6H", "INCHI", "c1ccccc1"),
            (_BENZENE_MOLBLOCK, "MOLBLOCK", "c1ccccc1"),
        ),
    )
    def test_mol_from_compound(self, value, identifier_type, expected):
        compound = reaction_pb2.Compound()
        compound.identifiers.add(value=value, type=identifier_type)
        mol = message_helpers.mol_from_compound(compound)
        assert Chem.MolToSmiles(mol) == expected
        mol, identifier = message_helpers.mol_from_compound(compound, return_identifier=True)
        assert Chem.MolToSmiles(mol) == expected
        assert identifier == compound.identifiers[0]

    @pytest.mark.parametrize("value,identifier_type", (("invalid_smiles", "SMILES"), ("invalid_inchi", "INCHI")))
    def test_mol_from_compound_failures(self, value, identifier_type):
        compound = reaction_pb2.Compound()
        compound.identifiers.add(value=value, type=identifier_type)
        with pytest.raises(ValueError, match="invalid structural identifier"):
            message_helpers.mol_from_compound(compound)

    def test_get_product_yield(self):
        product = reaction_pb2.ProductCompound()
        assert message_helpers.get_product_yield(product) is None
        product.measurements.add(type="YIELD", percentage=dict(value=23))
        assert 23 == message_helpers.get_product_yield(product)

    def test_check_compound_identifiers(self):
        compound = reaction_pb2.Compound()
        compound.identifiers.add(value="c1ccccc1", type="SMILES")
        message_helpers.check_compound_identifiers(compound)
        compound.identifiers.add(value="InChI=1S/C6H6/c1-2-4-6-5-3-1/h1-6H", type="INCHI")
        message_helpers.check_compound_identifiers(compound)
        compound.identifiers.add(value="c1ccc(O)cc1", type="SMILES")
        with pytest.raises(ValueError, match="inconsistent"):
            message_helpers.check_compound_identifiers(compound)

    def test_get_reaction_smiles(self):
        reaction = reaction_pb2.Reaction()
        reactant1 = reaction.inputs["reactant1"]
        reactant1.components.add(reaction_role="REACTANT").identifiers.add(value="c1ccccc1", type="SMILES")
        reactant1.components.add(reaction_role="SOLVENT").identifiers.add(value="N", type="SMILES")
        assert message_helpers.get_reaction_smiles(reaction, generate_if_missing=True) == "c1ccccc1>N>"
        reactant2 = reaction.inputs["reactant2"]
        reactant2.components.add(reaction_role="REACTANT").identifiers.add(value="Cc1ccccc1", type="SMILES")
        reactant2.components.add(reaction_role="SOLVENT").identifiers.add(value="N", type="SMILES")
        reaction.outcomes.add().products.add(reaction_role="PRODUCT").identifiers.add(value="O=C=O", type="SMILES")
        assert message_helpers.get_reaction_smiles(reaction, generate_if_missing=True) == "Cc1ccccc1.c1ccccc1>N>O=C=O"

    def test_get_reaction_smiles_failure(self):
        reaction = reaction_pb2.Reaction()
        reactant = reaction.inputs["reactant"]
        component = reactant.components.add(reaction_role="REACTANT")
        component.identifiers.add(value="benzene", type="NAME")
        with pytest.raises(ValueError, match="no valid reactants or products"):
            message_helpers.get_reaction_smiles(reaction, generate_if_missing=True)
        component.identifiers.add(value="c1ccccc1", type="SMILES")
        with pytest.raises(ValueError, match="must contain at least one"):
            message_helpers.get_reaction_smiles(reaction, generate_if_missing=True, allow_incomplete=False)
        reaction.outcomes.add().products.add(reaction_role="PRODUCT").identifiers.add(value="invalid", type="SMILES")
        with pytest.raises(ValueError, match="bad reaction SMILES"):
            message_helpers.get_reaction_smiles(reaction, generate_if_missing=True, allow_incomplete=False)

    def test_reaction_from_smiles(self):
        reaction_smiles = "[C:1].N>O>F.Cl"
        expected = reaction_pb2.Reaction()
        expected.identifiers.add(value=reaction_smiles, type="REACTION_SMILES")
        this_input = expected.inputs["from_reaction_smiles"]
        c_component = this_input.components.add()
        c_component.identifiers.add(value="C", type="SMILES")
        c_component.reaction_role = reaction_pb2.ReactionRole.REACTANT
        n_component = this_input.components.add()
        n_component.identifiers.add(value="N", type="SMILES")
        n_component.reaction_role = reaction_pb2.ReactionRole.REACTANT
        o_component = this_input.components.add()
        o_component.identifiers.add(value="O", type="SMILES")
        o_component.reaction_role = reaction_pb2.ReactionRole.REAGENT
        outcome = expected.outcomes.add()
        f_component = outcome.products.add()
        f_component.identifiers.add(value="F", type="SMILES")
        f_component.reaction_role = reaction_pb2.ReactionRole.PRODUCT
        cl_component = outcome.products.add()
        cl_component.identifiers.add(value="Cl", type="SMILES")
        cl_component.reaction_role = reaction_pb2.ReactionRole.PRODUCT
        assert message_helpers.reaction_from_smiles(reaction_smiles) == expected

    @pytest.mark.parametrize(
        "doi,expected",
        (
            (
                "https://dx.doi.org/10.1021/acscatal.0c02247",
                "10.1021/acscatal.0c02247",
            ),
        ),
    )
    def test_parse_doi(self, doi, expected):
        assert message_helpers.parse_doi(doi) == expected

    def test_fetch_dataset(self):
        dataset = message_helpers.fetch_dataset("ord_dataset-35a5a513f1dd44a3a97c88da99f81a00")
        assert len(dataset.reactions) == 7


class TestFindSubmessages:
    def test_scalar(self):
        message = test_pb2.Scalar(int32_value=5, float_value=6.7)
        assert len(message_helpers.find_submessages(message, test_pb2.Scalar)) == 0
        with pytest.raises(TypeError, match="must be a Protocol Buffer"):
            message_helpers.find_submessages(message, float)

    def test_nested(self):
        message = test_pb2.Nested()
        assert len(message_helpers.find_submessages(message, test_pb2.Nested.Child)) == 0
        message.child.value = 5.6
        submessages = message_helpers.find_submessages(message, test_pb2.Nested.Child)
        assert len(submessages) == 1
        # Show that the returned submessages work as references.
        submessages[0].value = 7.8
        assert round(abs(message.child.value - 7.8), 4) == 0

    def test_repeated_nested(self):
        message = test_pb2.RepeatedNested()
        message.children.add().value = 1.2
        message.children.add().value = 3.4
        assert len(message_helpers.find_submessages(message, test_pb2.RepeatedNested.Child)) == 2

    def test_map_nested(self):
        message = test_pb2.MapNested()
        message.children["one"].value = 1.2
        message.children["two"].value = 3.4
        assert len(message_helpers.find_submessages(message, test_pb2.MapNested.Child)) == 2

    def test_compounds(self):
        message = reaction_pb2.Reaction()
        message.inputs["test"].components.add().identifiers.add(type="NAME", value="aspirin")
        assert len(message_helpers.find_submessages(message, reaction_pb2.Compound)) == 1


class TestBuildData:
    def test_build_data(self, tmp_path):
        data = b"test data"
        filename = (tmp_path / "test.data").as_posix()
        with open(filename, "wb") as f:
            f.write(data)
        message = message_helpers.build_data(filename, description="binary data")
        assert message.bytes_value == data
        assert message.description == "binary data"
        assert message.format == "data"

    def test_bad_filename(self):
        with pytest.raises(ValueError, match="cannot deduce the file format"):
            message_helpers.build_data("testdata", "no description")


class TestBuildCompound:
    def test_smiles_and_name(self):
        compound = message_helpers.build_compound(smiles="c1ccccc1", name="benzene")
        expected = reaction_pb2.Compound(
            identifiers=[
                reaction_pb2.CompoundIdentifier(value="c1ccccc1", type="SMILES"),
                reaction_pb2.CompoundIdentifier(value="benzene", type="NAME"),
            ]
        )
        assert compound == expected

    def test_unmeasured_amount(self):
        compound = message_helpers.build_compound(smiles="CCO", amount="catalytic")
        expected = reaction_pb2.Compound(
            identifiers=[
                reaction_pb2.CompoundIdentifier(value="CCO", type="SMILES"),
            ],
            amount=dict(unmeasured=dict(type="CATALYTIC")),
        )
        assert compound == expected

    @pytest.mark.parametrize(
        "amount,expected",
        (
            ("1.2 g", reaction_pb2.Mass(value=1.2, units="GRAM")),
            ("3.4 mol", reaction_pb2.Moles(value=3.4, units="MOLE")),
            ("5.6 mL", reaction_pb2.Volume(value=5.6, units="MILLILITER")),
        ),
    )
    def test_amount(self, amount, expected):
        compound = message_helpers.build_compound(amount=amount)
        assert getattr(compound.amount, compound.amount.WhichOneof("kind")) == expected

    @pytest.mark.parametrize("amount", ("1.2", "-3.4 g"))
    def test_bad_amount(self, amount):
        with pytest.raises((KeyError, ValueError)):
            message_helpers.build_compound(amount=amount)

    def test_role(self):
        compound = message_helpers.build_compound(role="solvent")
        assert compound.reaction_role == reaction_pb2.ReactionRole.SOLVENT

    def test_bad_role(self):
        with pytest.raises(KeyError, match="not a supported type"):
            message_helpers.build_compound(role="flavorant")

    def test_is_limiting(self):
        assert message_helpers.build_compound(is_limiting=True).is_limiting
        assert not message_helpers.build_compound(is_limiting=False).is_limiting
        assert not message_helpers.build_compound().HasField("is_limiting")

    @pytest.mark.parametrize(
        "prep,details,expected",
        (
            ("dried", None, reaction_pb2.CompoundPreparation(type="DRIED")),
            (
                "dried",
                "in the fire of the sun",
                reaction_pb2.CompoundPreparation(type="DRIED", details="in the fire of the sun"),
            ),
            (
                "custom",
                "threw it on the ground",
                reaction_pb2.CompoundPreparation(type="CUSTOM", details="threw it on the ground"),
            ),
        ),
    )
    def test_prep(self, prep, details, expected):
        compound = message_helpers.build_compound(prep=prep, prep_details=details)
        assert compound.preparations[0] == expected

    def test_bad_prep(self):
        with pytest.raises(KeyError, match="not a supported type"):
            message_helpers.build_compound(prep="shaken")

    def test_prep_details_without_prep(self):
        with pytest.raises(ValueError, match="prep must be provided"):
            message_helpers.build_compound(prep_details="rinsed gently")

    def test_custom_prep_without_details(self):
        with pytest.raises(ValueError, match="prep_details must be provided"):
            message_helpers.build_compound(prep="custom")

    def test_vendor(self):
        assert message_helpers.build_compound(vendor="Sally").source.vendor == "Sally"


class TestSetSoluteMoles:
    def test_set_solute_moles_should_fail(self):
        solute = message_helpers.build_compound(name="Solute")
        solvent = message_helpers.build_compound(name="Solvent")
        with pytest.raises(ValueError, match="defined volume"):
            message_helpers.set_solute_moles(solute, [solvent], "10 mM")

        solute = message_helpers.build_compound(name="Solute", amount="1 mol")
        solvent = message_helpers.build_compound(name="Solvent", amount="1 L")
        with pytest.raises(ValueError, match="overwrite"):
            message_helpers.set_solute_moles(solute, [solvent], "10 mM")

    def test_set_solute_moles(self):
        solute = message_helpers.build_compound(name="Solute")
        solvent2 = message_helpers.build_compound(name="Solvent", amount="100 mL")
        message_helpers.set_solute_moles(solute, [solvent2], "1 molar")
        assert solute.amount.moles == reaction_pb2.Moles(units="MILLIMOLE", value=100)
        solvent3 = message_helpers.build_compound(name="Solvent", amount="75 uL")
        message_helpers.set_solute_moles(solute, [solvent3], "3 mM", overwrite=True)
        assert solute.amount.moles.units == reaction_pb2.Moles.NANOMOLE
        assert round(abs(solute.amount.moles.value - 225), 4) == 0
        solvent4 = message_helpers.build_compound(name="Solvent", amount="0.2 uL")
        message_helpers.set_solute_moles(solute, [solvent4], "30 mM", overwrite=True)
        assert solute.amount.moles == reaction_pb2.Moles(units="NANOMOLE", value=6)
        solvent5 = message_helpers.build_compound(name="Solvent", amount="0.8 uL")
        message_helpers.set_solute_moles(solute, [solvent4, solvent5], "30 mM", overwrite=True)
        assert solute.amount.moles == reaction_pb2.Moles(units="NANOMOLE", value=30)


class TestCompoundIdentifiers:
    def test_identifier_setters(self):
        compound = reaction_pb2.Compound()
        identifier = message_helpers.set_compound_name(compound, "water")
        assert identifier == reaction_pb2.CompoundIdentifier(type="NAME", value="water")
        assert compound.identifiers[0] == reaction_pb2.CompoundIdentifier(type="NAME", value="water")
        message_helpers.set_compound_smiles(compound, "O")
        assert compound.identifiers[1] == reaction_pb2.CompoundIdentifier(type="SMILES", value="O")
        identifier = message_helpers.set_compound_name(compound, "ice")
        assert identifier == reaction_pb2.CompoundIdentifier(type="NAME", value="ice")
        assert compound.identifiers[0] == reaction_pb2.CompoundIdentifier(type="NAME", value="ice")
        compound = reaction_pb2.Compound()
        _ = message_helpers.set_compound_molblock(compound, _BENZENE_MOLBLOCK)
        assert _BENZENE_MOLBLOCK == compound.identifiers[0].value

    def test_identifier_getters(self):
        compound = reaction_pb2.Compound()
        compound.identifiers.add(type="NAME", value="water")
        assert message_helpers.get_compound_name(compound) == "water"
        assert message_helpers.get_compound_smiles(compound) is None
        compound.identifiers.add(type="SMILES", value="O")
        assert message_helpers.get_compound_smiles(compound) == "O"
        assert message_helpers.smiles_from_compound(compound) == "O"
        compound = reaction_pb2.Compound()
        compound.identifiers.add(type="MOLBLOCK", value=_BENZENE_MOLBLOCK)
        assert message_helpers.get_compound_molblock(compound) == _BENZENE_MOLBLOCK
        assert message_helpers.molblock_from_compound(compound) == _BENZENE_MOLBLOCK


class TestSetDativeBonds:
    def test_has_transition_metal(self):
        assert not message_helpers.has_transition_metal(Chem.MolFromSmiles("P"))
        assert message_helpers.has_transition_metal(Chem.MolFromSmiles("Cl[Pd]Cl"))

    @pytest.mark.parametrize(
        "smiles,from_atoms,expected", (("[PH3][Pd](Cl)(Cl)[NH3]", ("N", "P"), "N->[Pd](<-P)(Cl)Cl"),)
    )
    def test_set_dative_bonds(self, smiles, from_atoms, expected):
        mol = Chem.MolFromSmiles(smiles, sanitize=False)
        dative_mol = message_helpers.set_dative_bonds(mol, from_atoms=from_atoms)
        assert Chem.MolToSmiles(dative_mol) == expected


class TestLoadAndWriteMessage:
    @pytest.fixture
    def messages(self) -> list:
        yield [
            test_pb2.Scalar(int32_value=3, float_value=4.5),
            test_pb2.RepeatedScalar(values=[1.2, 3.4]),
            test_pb2.Enum(value="FIRST"),
            test_pb2.RepeatedEnum(values=["FIRST", "SECOND"]),
            test_pb2.Nested(child=test_pb2.Nested.Child(value=1.2)),
        ]

    @pytest.mark.parametrize("suffix", (".pbtxt", ".pb", ".json", ".pbtxt.gz", ".pb.gz", ".json.gz"))
    def test_round_trip(self, suffix, messages):
        for message in messages:
            with tempfile.NamedTemporaryFile(suffix=suffix) as f:
                message_helpers.write_message(message, f.name)
                f.flush()
                assert message == message_helpers.load_message(f.name, type(message))

    def test_gzip_reproducibility(self, messages, tmp_path):
        filename = (tmp_path / "test.pb.gz").as_posix()
        for message in messages:
            message_helpers.write_message(message, filename)
            with open(filename, "rb") as f:
                value = f.read()
            time.sleep(1)
            message_helpers.write_message(message, filename)
            with open(filename, "rb") as f:
                assert f.read() == value

    def test_bad_binary(self):
        with tempfile.NamedTemporaryFile(suffix=".pb") as f:
            message = test_pb2.RepeatedScalar(values=[1.2, 3.4])
            f.write(message.SerializeToString())
            f.flush()
            # NOTE(kearnes): The decoder is not perfect; for example, it will
            # not be able to distinguish from a message with the same tags and
            # types (e.g. test_pb2.Scalar and test_pb2.RepeatedScalar).
            with pytest.raises(ValueError, match="parsing"):
                message_helpers.load_message(f.name, test_pb2.Nested)

    def test_bad_json(self):
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".json") as f:
            message = test_pb2.RepeatedScalar(values=[1.2, 3.4])
            f.write(json_format.MessageToJson(message))
            f.flush()
            with pytest.raises(ValueError, match='no field named "values"'):
                message_helpers.load_message(f.name, test_pb2.Nested)

    def test_bad_pbtxt(self):
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".pbtxt") as f:
            message = test_pb2.RepeatedScalar(values=[1.2, 3.4])
            f.write(text_format.MessageToString(message))
            f.flush()
            with pytest.raises(ValueError, match='no field named "values"'):
                message_helpers.load_message(f.name, test_pb2.Nested)

    def test_bad_suffix(self):
        message = test_pb2.RepeatedScalar(values=[1.2, 3.4])
        with pytest.raises(ValueError, match="not a valid MessageFormat"):
            message_helpers.write_message(message, "test.proto")


class TestCreateMessage:
    @pytest.mark.parametrize(
        "message_name,expected_class",
        (
            ("Reaction", reaction_pb2.Reaction),
            ("Temperature", reaction_pb2.Temperature),
            ("TemperatureConditions.Measurement", reaction_pb2.TemperatureConditions.Measurement),
        ),
    )
    def test_valid_messages(self, message_name, expected_class):
        message = message_helpers.create_message(message_name)
        assert isinstance(message, expected_class)

    @pytest.mark.parametrize("message_name", ("reaction", "aosdifjasdf"))
    def test_invalid_messages(self, message_name):
        with pytest.raises(ValueError, match="Cannot resolve"):
            message_helpers.create_message(message_name)


class TestMessagesToDataFrame:
    @pytest.mark.parametrize(
        "message,expected",
        (
            (test_pb2.Scalar(int32_value=3, float_value=4.5), {"int32_value": 3, "float_value": 4.5}),
            (
                test_pb2.Scalar(int32_value=0, float_value=0.0, bool_value=False),
                {"float_value": 0.0, "bool_value": False},
            ),
            (test_pb2.RepeatedScalar(values=[1.2, 3.4]), {"values[0]": 1.2, "values[1]": 3.4}),
            (test_pb2.Enum(value="FIRST"), {"value": "FIRST"}),
            (test_pb2.RepeatedEnum(values=["FIRST", "SECOND"]), {"values[0]": "FIRST", "values[1]": "SECOND"}),
            (test_pb2.Nested(child=test_pb2.Nested.Child(value=1.2)), {"child.value": 1.2}),
            (
                test_pb2.RepeatedNested(
                    children=[test_pb2.RepeatedNested.Child(value=1.2), test_pb2.RepeatedNested.Child(value=3.4)]
                ),
                {"children[0].value": 1.2, "children[1].value": 3.4},
            ),
            (test_pb2.Map(values={"a": 1.2, "b": 3.4}), {'values["a"]': 1.2, 'values["b"]': 3.4}),
            (
                test_pb2.MapNested(
                    children={"a": test_pb2.MapNested.Child(value=1.2), "b": test_pb2.MapNested.Child(value=3.4)}
                ),
                {'children["a"].value': 1.2, 'children["b"].value': 3.4},
            ),
        ),
    )
    def test_message_to_row(self, message, expected):
        row = message_helpers.message_to_row(message)
        pd.testing.assert_frame_equal(pd.DataFrame([row]), pd.DataFrame([expected]), check_like=True)

    def test_messages_to_dataframe(self):
        reaction1 = reaction_pb2.Reaction()
        input_test = reaction1.inputs["test"]
        outcome = reaction1.outcomes.add()
        component = input_test.components.add()
        component.identifiers.add(type="SMILES", value="CCO")
        component.is_limiting = True
        component.amount.mass.value = 1.2
        component.amount.mass.units = reaction_pb2.Mass.GRAM
        outcome.conversion.value = 3.4
        outcome.conversion.precision = 5.6
        reaction2 = reaction_pb2.Reaction()
        input_test = reaction2.inputs["test"]
        outcome = reaction2.outcomes.add()
        component = input_test.components.add()
        component.identifiers.add(type="SMILES", value="CCC")
        component.is_limiting = False
        component.amount.mass.value = 7.8
        component.amount.mass.units = reaction_pb2.Mass.GRAM
        outcome.conversion.value = 9.1
        outcome.conversion.precision = 2.3
        expected = pd.DataFrame(
            {
                'inputs["test"].components[0].identifiers[0].type': "SMILES",
                'inputs["test"].components[0].identifiers[0].value': ["CCO", "CCC"],
                'inputs["test"].components[0].is_limiting': [True, False],
                'inputs["test"].components[0].amount.mass.value': [1.2, 7.8],
                'inputs["test"].components[0].amount.mass.units': "GRAM",
                "outcomes[0].conversion.value": [3.4, 9.1],
                "outcomes[0].conversion.precision": [5.6, 2.3],
            }
        )
        pd.testing.assert_frame_equal(
            message_helpers.messages_to_dataframe([reaction1, reaction2]),
            expected,
            check_like=True,
        )
        # Drop constant columns and test again with drop_constant_columns=True.
        del expected['inputs["test"].components[0].identifiers[0].type']
        del expected['inputs["test"].components[0].amount.mass.units']
        pd.testing.assert_frame_equal(
            message_helpers.messages_to_dataframe([reaction1, reaction2], drop_constant_columns=True),
            expected,
            check_like=True,
        )



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Table mappings for Reaction protos.

Notes:
    * Foreign keys to the `reaction` table are done using the `id` column, not the ORD reaction ID (`reaction_id`).
      However, the `reaction_id` column is used when the reaction ID is specifically called for, as with crude inputs.
    * When a message type is used in multiple places, use Joined Table Inheritance; see
      https://docs.sqlalchemy.org/en/14/orm/inheritance.html#joined-table-inheritance.
    * The naming might be a bit confusing: classes for single-use and parent protos have the same name
      as the corresponding message, while classes for multi-use protos (child classes) are named as
      _<ContainingClass><Attribute>. For example, Reaction.identifiers uses the ReactionIdentifier class,
      while Reaction.inputs uses the _ReactionInputs class (since _ReactionInput is used in multiple places).
      The effect is that all tables storing proto information are named after their corresponding proto
      messages, with extra child tables for storing relationships.
    * Parent and child class names are prepended with underscores; e.g. _ReactionInput. Every parent class has an
      associated with_polymorphic wrapper for query construction, named without the leading underscore;
      e.g. ReactionInput.
    * Only message types are allowed for repeated/mapped values in the ORM (not scalar types). Specifically:
        * MassSpecMeasurementDetails.eic_masses is converted from repeated float to repeated FloatValue.
        * Dataset.reaction_ids is converted from repeated string to repeated DatasetId.
      # TODO(skearnes): Update the proto definition?
    * Source.id was renamed in the ORM to Source.vendor_id to avoid conflicting with the `id` column.
      # TODO(skearnes): Update the proto definition?
"""
from __future__ import annotations

from inspect import getmro
from typing import Mapping, Optional, Type

from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.message import Message
from sqlalchemy import Boolean, Column, Enum, Float, Integer, ForeignKey, LargeBinary, Text
from sqlalchemy.orm import relationship, with_polymorphic

import ord_schema.orm.structure  # pylint: disable=unused-import
from ord_schema import message_helpers
from ord_schema.orm import Base, Child, Parent
from ord_schema.proto import dataset_pb2
from ord_schema.proto import reaction_pb2

# pylint:disable=missing-class-docstring


class Dataset(Base):
    name = Column(Text)
    description = Column(Text)
    reactions = relationship("Reaction")
    reaction_ids = relationship("ReactionId")
    dataset_id = Column(Text, nullable=False, unique=True)


class ReactionId(Base):
    dataset_id = Column(Integer, ForeignKey("dataset.id", ondelete="CASCADE"), nullable=False)

    reaction_id = Column(Text, ForeignKey("reaction.reaction_id", ondelete="CASCADE"), nullable=False)


class DatasetExample(Base):
    dataset_id = Column(Text, ForeignKey("dataset.dataset_id", ondelete="CASCADE"), nullable=False)
    description = Column(Text)
    url = Column(Text)
    created = relationship("_DatasetExampleCreated", uselist=False)


class Reaction(Base):
    dataset_id = Column(Integer, ForeignKey("dataset.id", ondelete="CASCADE"), nullable=False)
    proto = Column(LargeBinary, nullable=False)

    identifiers = relationship("ReactionIdentifier")
    inputs = relationship("_ReactionInputs")
    setup = relationship("ReactionSetup", uselist=False)
    conditions = relationship("ReactionConditions", uselist=False)
    notes = relationship("ReactionNotes", uselist=False)
    observations = relationship("ReactionObservation")
    workups = relationship("ReactionWorkup")
    outcomes = relationship("ReactionOutcome")
    provenance = relationship("ReactionProvenance", uselist=False)
    reaction_id = Column(Text, nullable=False, unique=True)


class ReactionIdentifier(Base):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False)

    type = Column(
        Enum(*reaction_pb2.ReactionIdentifier.IdentifierType.keys(), name="ReactionIdentifier.IdentifierType")
    )
    details = Column(Text)
    value = Column(Text)
    is_mapped = Column(Boolean)


class _ReactionInput(Parent, Base):
    components = relationship("_ReactionInputComponents")
    crude_components = relationship("CrudeComponent")
    addition_order = Column(Integer)
    addition_time = relationship("_ReactionInputAdditionTime", uselist=False)
    addition_speed = relationship("AdditionSpeed", uselist=False)
    addition_duration = relationship("_ReactionInputAdditionDuration", uselist=False)
    flow_rate = relationship("FlowRate", uselist=False)
    addition_device = relationship("AdditionDevice", uselist=False)
    addition_temperature = relationship("_ReactionInputAdditionTemperature", uselist=False)


class _ReactionInputs(Child, _ReactionInput):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False)
    key = Column(Text)  # Map key.


class _ReactionWorkupInput(Child, _ReactionInput):
    reaction_workup_id = Column(
        Integer, ForeignKey("reaction_workup.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class AdditionSpeed(Base):
    reaction_input_id = Column(
        Integer, ForeignKey("reaction_input.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.ReactionInput.AdditionSpeed.AdditionSpeedType.keys(),
            name="ReactionInput.AdditionSpeed.AdditionSpeedType",
        )
    )
    details = Column(Text)


class AdditionDevice(Base):
    reaction_input_id = Column(
        Integer, ForeignKey("reaction_input.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.ReactionInput.AdditionDevice.AdditionDeviceType.keys(),
            name="ReactionInput.AdditionDevice.AdditionDeviceType",
        )
    )
    details = Column(Text)


class _Amount(Parent, Base):
    mass = relationship("Mass", uselist=False)
    moles = relationship("Moles", uselist=False)
    volume = relationship("_AmountVolume", uselist=False)
    unmeasured = relationship("UnmeasuredAmount", uselist=False)
    volume_includes_solutes = Column(Boolean)


class _CrudeComponentAmount(Child, _Amount):
    crude_component_id = Column(Integer, ForeignKey("crude_component.id", ondelete="CASCADE"), nullable=False)


class _CompoundAmount(Child, _Amount):
    compound_id = Column(Integer, ForeignKey("compound.id", ondelete="CASCADE"), nullable=False)


class _ReactionWorkupAmount(Child, _Amount):
    reaction_workup_id = Column(Integer, ForeignKey("reaction_workup.id", ondelete="CASCADE"), nullable=False)


class _ProductMeasurementAmount(Child, _Amount):
    product_measurement_id = Column(Integer, ForeignKey("product_measurement.id", ondelete="CASCADE"), nullable=False)


class UnmeasuredAmount(Base):
    amount_id = Column(Integer, ForeignKey("amount.id", ondelete="CASCADE"), nullable=False, unique=True)

    type = Column(
        Enum(*reaction_pb2.UnmeasuredAmount.UnmeasuredAmountType.keys(), name="UnmeasuredAmount.UnmeasuredAmountType")
    )
    details = Column(Text)


class CrudeComponent(Base):
    reaction_input_id = Column(Integer, ForeignKey("reaction_input.id", ondelete="CASCADE"), nullable=False)

    reaction_id = Column(Text, ForeignKey("reaction.reaction_id", ondelete="CASCADE"), nullable=False)
    includes_workup = Column(Boolean)
    has_derived_amount = Column(Boolean)
    amount = relationship("_CrudeComponentAmount", uselist=False)


# Shared enum; see https://docs.sqlalchemy.org/en/14/dialects/postgresql.html#sqlalchemy.dialects.postgresql.ENUM.
ReactionRoleType = Enum(
    *reaction_pb2.ReactionRole.ReactionRoleType.keys(), name="ReactionRole.ReactionRoleType", metadata=Base.metadata
)


class _Compound(Parent, Base):
    identifiers = relationship("_CompoundIdentifiers")
    amount = relationship("_CompoundAmount", uselist=False)
    reaction_role = Column(ReactionRoleType)
    is_limiting = Column(Boolean)
    preparations = relationship("CompoundPreparation")
    source = relationship("Source", uselist=False)
    features = relationship("_CompoundFeatures")
    analyses = relationship("_CompoundAnalyses")
    structure = relationship("_CompoundStructure", uselist=False)


class _ReactionInputComponents(Child, _Compound):
    reaction_input_id = Column(Integer, ForeignKey("reaction_input.id", ondelete="CASCADE"), nullable=False)


class _ProductMeasurementAuthenticStandard(Child, _Compound):
    product_measurement_id = Column(
        Integer, ForeignKey("product_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class Source(Base):
    compound_id = Column(Integer, ForeignKey("compound.id", ondelete="CASCADE"), nullable=False, unique=True)

    vendor = Column(Text)
    vendor_id = Column(Text)
    lot = Column(Text)


class CompoundPreparation(Base):
    compound_id = Column(Integer, ForeignKey("compound.id", ondelete="CASCADE"), nullable=False)

    type = Column(
        Enum(*reaction_pb2.CompoundPreparation.PreparationType.keys(), name="CompoundPreparation.PreparationType")
    )
    details = Column(Text)
    reaction_id = Column(Text, ForeignKey("reaction.reaction_id", ondelete="CASCADE"))


class _CompoundIdentifier(Parent, Base):
    type = Column(
        Enum(*reaction_pb2.CompoundIdentifier.IdentifierType.keys(), name="CompoundIdentifier.IdentifierType")
    )
    details = Column(Text)
    value = Column(Text)


class _CompoundIdentifiers(Child, _CompoundIdentifier):
    compound_id = Column(Integer, ForeignKey("compound.id", ondelete="CASCADE"), nullable=False)


class _ProductCompoundIdentifiers(Child, _CompoundIdentifier):
    product_compound_id = Column(Integer, ForeignKey("product_compound.id", ondelete="CASCADE"), nullable=False)


class Vessel(Base):
    reaction_setup_id = Column(
        Integer, ForeignKey("reaction_setup.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(Enum(*reaction_pb2.Vessel.VesselType.keys(), name="Vessel.VesselType"))
    details = Column(Text)
    material = relationship("VesselMaterial", uselist=False)
    preparations = relationship("VesselPreparation")
    attachments = relationship("VesselAttachment")
    volume = relationship("_VesselVolume", uselist=False)
    plate_id = Column(Text)
    plate_position = Column(Text)


class VesselMaterial(Base):
    vessel_id = Column(Integer, ForeignKey("vessel.id", ondelete="CASCADE"), nullable=False, unique=True)

    type = Column(
        Enum(*reaction_pb2.VesselMaterial.VesselMaterialType.keys(), name="VesselMaterial.VesselMaterialType")
    )
    details = Column(Text)


class VesselAttachment(Base):
    vessel_id = Column(Integer, ForeignKey("vessel.id", ondelete="CASCADE"), nullable=False)

    type = Column(
        Enum(*reaction_pb2.VesselAttachment.VesselAttachmentType.keys(), name="VesselAttachment.VesselAttachmentType")
    )
    details = Column(Text)


class VesselPreparation(Base):
    vessel_id = Column(Integer, ForeignKey("vessel.id", ondelete="CASCADE"), nullable=False)

    type = Column(
        Enum(
            *reaction_pb2.VesselPreparation.VesselPreparationType.keys(), name="VesselPreparation.VesselPreparationType"
        )
    )
    details = Column(Text)


class ReactionSetup(Base):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False, unique=True)

    vessel = relationship("Vessel", uselist=False)
    is_automated = Column(Boolean)
    automation_platform = Column(Text)
    automation_code = relationship("_ReactionSetupAutomationCode")
    environment = relationship("ReactionEnvironment", uselist=False)


class ReactionEnvironment(Base):
    reaction_setup_id = Column(
        Integer, ForeignKey("reaction_setup.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.ReactionSetup.ReactionEnvironment.ReactionEnvironmentType.keys(),
            name="ReactionSetup.ReactionEnvironment.ReactionEnvironmentType",
        )
    )
    details = Column(Text)


class ReactionConditions(Base):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False, unique=True)

    temperature = relationship("_ReactionConditionsTemperature", uselist=False)
    pressure = relationship("PressureConditions", uselist=False)
    stirring = relationship("_ReactionConditionsStirring", uselist=False)
    illumination = relationship("IlluminationConditions", uselist=False)
    electrochemistry = relationship("ElectrochemistryConditions", uselist=False)
    flow = relationship("FlowConditions", uselist=False)
    reflux = Column(Boolean)
    ph = Column(Float)
    conditions_are_dynamic = Column(Boolean)
    details = Column(Text)


class _TemperatureConditions(Parent, Base):
    control = relationship("TemperatureControl", uselist=False)
    setpoint = relationship("_TemperatureConditionsSetpoint", uselist=False)
    measurements = relationship("TemperatureConditionsMeasurement")


class _ReactionConditionsTemperature(Child, _TemperatureConditions):
    reaction_conditions_id = Column(
        Integer, ForeignKey("reaction_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ReactionWorkupTemperature(Child, _TemperatureConditions):
    reaction_workup_id = Column(
        Integer, ForeignKey("reaction_workup.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class TemperatureControl(Base):
    temperature_conditions_id = Column(
        Integer, ForeignKey("temperature_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.TemperatureConditions.TemperatureControl.TemperatureControlType.keys(),
            name="TemperatureConditions.TemperatureControl.TemperatureControlType",
        )
    )
    details = Column(Text)


class TemperatureConditionsMeasurement(Base):
    temperature_conditions_id = Column(
        Integer, ForeignKey("temperature_conditions.id", ondelete="CASCADE"), nullable=False
    )

    type = Column(
        Enum(
            *reaction_pb2.TemperatureConditions.Measurement.MeasurementType.keys(),
            name="TemperatureConditions.Measurement.MeasurementType",
        )
    )
    details = Column(Text)
    time = relationship("_TemperatureConditionsMeasurementTime", uselist=False)
    temperature = relationship("_TemperatureConditionsMeasurementTemperature", uselist=False)


class PressureConditions(Base):
    reaction_conditions_id = Column(
        Integer, ForeignKey("reaction_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    control = relationship("PressureControl", uselist=False)
    setpoint = relationship("_PressureConditionsSetpoint", uselist=False)
    atmosphere = relationship("Atmosphere", uselist=False)
    measurements = relationship("PressureConditionsMeasurement")


class PressureControl(Base):
    pressure_conditions_id = Column(
        Integer, ForeignKey("pressure_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.PressureConditions.PressureControl.PressureControlType.keys(),
            name="PressureConditions.PressureControl.PressureControlType",
        )
    )
    details = Column(Text)


class Atmosphere(Base):
    pressure_conditionss_id = Column(
        Integer, ForeignKey("pressure_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.PressureConditions.Atmosphere.AtmosphereType.keys(),
            name="PressureConditions.Atmosphere.AtmosphereType",
        )
    )
    details = Column(Text)


class PressureConditionsMeasurement(Base):
    pressure_conditionss_id = Column(Integer, ForeignKey("pressure_conditions.id", ondelete="CASCADE"), nullable=False)

    type = Column(
        Enum(
            *reaction_pb2.PressureConditions.Measurement.MeasurementType.keys(),
            name="PressureConditions.Measurement.MeasurementType",
        )
    )
    details = Column(Text)
    time = relationship("_PressureConditionsMeasurementTime", uselist=False)
    pressure = relationship("_PressureConditionsMeasurementPressure", uselist=False)


class _StirringConditions(Parent, Base):
    type = Column(
        Enum(*reaction_pb2.StirringConditions.StirringMethodType.keys(), name="StirringConditions.StirringMethodType")
    )
    details = Column(Text)
    rate = relationship("StirringRate", uselist=False)


class _ReactionConditionsStirring(Child, _StirringConditions):
    reaction_conditions_id = Column(
        Integer, ForeignKey("reaction_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ReactionWorkupStirring(Child, _StirringConditions):
    reaction_workup_id = Column(
        Integer, ForeignKey("reaction_workup.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class StirringRate(Base):
    stirring_conditions_id = Column(
        Integer, ForeignKey("stirring_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.StirringConditions.StirringRate.StirringRateType.keys(),
            name="StirringConditions.StirringRate.StirringRateType",
        )
    )
    details = Column(Text)
    rpm = Column(Integer)


class IlluminationConditions(Base):
    reaction_conditions_id = Column(
        Integer, ForeignKey("reaction_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.IlluminationConditions.IlluminationType.keys(), name="IlluminationConditions.IlluminationType"
        )
    )
    details = Column(Text)
    peak_wavelength = relationship("_IlluminationConditionsPeakWavelength", uselist=False)
    color = Column(Text)
    distance_to_vessel = relationship("_IlluminationConditionsDistanceToVessel", uselist=False)


class ElectrochemistryConditions(Base):
    reaction_conditions_id = Column(
        Integer, ForeignKey("reaction_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.ElectrochemistryConditions.ElectrochemistryType.keys(),
            name="ElectrochemistryConditions.ElectrochemistryType",
        )
    )
    details = Column(Text)
    current = relationship("_ElectrochemistryConditionsCurrent", uselist=False)
    voltage = relationship("_ElectrochemistryConditionsVoltage", uselist=False)
    anode_material = Column(Text)
    cathode_material = Column(Text)
    electrode_separation = relationship("_ElectrochemistryConditionsElectrodeSeparation", uselist=False)
    measurements = relationship("ElectrochemistryConditionsMeasurement")
    cell = relationship("ElectrochemistryCell", uselist=False)


class ElectrochemistryConditionsMeasurement(Base):
    electrochemistry_conditions_id = Column(
        Integer, ForeignKey("electrochemistry_conditions.id", ondelete="CASCADE"), nullable=False
    )

    time = relationship("_ElectrochemistryConditionsMeasurementTime", uselist=False)
    current = relationship("_ElectrochemistryConditionsMeasurementCurrent", uselist=False)
    voltage = relationship("_ElectrochemistryConditionsMeasurementVoltage", uselist=False)


class ElectrochemistryCell(Base):
    electrochemistry_conditions_id = Column(
        Integer, ForeignKey("electrochemistry_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.ElectrochemistryConditions.ElectrochemistryCell.ElectrochemistryCellType.keys(),
            name="ElectrochemistryConditions.ElectrochemistryCell.ElectrochemistryCellType",
        )
    )
    details = Column(Text)


class FlowConditions(Base):
    reaction_conditions_id = Column(
        Integer, ForeignKey("reaction_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(Enum(*reaction_pb2.FlowConditions.FlowType.keys(), name="FlowConditions.FlowType"))
    details = Column(Text)
    pump_type = Column(Text)
    tubing = relationship("Tubing", uselist=False)


class Tubing(Base):
    flow_conditions_id = Column(
        Integer, ForeignKey("flow_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(Enum(*reaction_pb2.FlowConditions.Tubing.TubingType.keys(), name="FlowConditions.Tubing.TubingType"))
    details = Column(Text)
    diameter = relationship("_TubingDiameter", uselist=False)


class ReactionNotes(Base):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False, unique=True)

    is_heterogeneous = Column(Boolean)
    forms_precipitate = Column(Boolean)
    is_exothermic = Column(Boolean)
    offgasses = Column(Boolean)
    is_sensitive_to_moisture = Column(Boolean)
    is_sensitive_to_oxygen = Column(Boolean)
    is_sensitive_to_light = Column(Boolean)
    safety_notes = Column(Text)
    procedure_details = Column(Text)


class ReactionObservation(Base):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False)

    time = relationship("_ReactionObservationTime", uselist=False)
    comment = Column(Text)
    image = relationship("_ReactionObservationImage", uselist=False)


class ReactionWorkup(Base):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False)

    type = Column(Enum(*reaction_pb2.ReactionWorkup.WorkupType.keys(), name="ReactionWorkup.WorkupType"))
    details = Column(Text)
    duration = relationship("_ReactionWorkupDuration", uselist=False)
    input = relationship("_ReactionWorkupInput", uselist=False)
    amount = relationship("_ReactionWorkupAmount", uselist=False)
    temperature = relationship("_ReactionWorkupTemperature", uselist=False)
    keep_phase = Column(Text)
    stirring = relationship("_ReactionWorkupStirring", uselist=False)
    target_ph = Column(Float)
    is_automated = Column(Boolean)


class ReactionOutcome(Base):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False)

    reaction_time = relationship("_ReactionOutcomeReactionTime", uselist=False)
    conversion = relationship("_ReactionOutcomeConversion", uselist=False)
    products = relationship("ProductCompound")
    analyses = relationship("_ReactionOutcomeAnalyses")


class ProductCompound(Base):
    reaction_outcome_id = Column(Integer, ForeignKey("reaction_outcome.id", ondelete="CASCADE"), nullable=False)

    identifiers = relationship("_ProductCompoundIdentifiers")
    is_desired_product = Column(Boolean)
    measurements = relationship("ProductMeasurement")
    isolated_color = Column(Text)
    texture = relationship("Texture", uselist=False)
    features = relationship("_ProductCompoundFeatures")
    reaction_role = Column(ReactionRoleType)
    structure = relationship("_ProductCompoundStructure", uselist=False)


class Texture(Base):
    product_compound_id = Column(
        Integer, ForeignKey("product_compound.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(*reaction_pb2.ProductCompound.Texture.TextureType.keys(), name="ProductCompound.Texture.TextureType")
    )
    details = Column(Text)


class ProductMeasurement(Base):
    product_compound_id = Column(Integer, ForeignKey("product_compound.id", ondelete="CASCADE"), nullable=False)

    analysis_key = Column(Text)
    type = Column(
        Enum(*reaction_pb2.ProductMeasurement.MeasurementType.keys(), name="ProductMeasurement.MeasurementType")
    )
    details = Column(Text)
    uses_internal_standard = Column(Boolean)
    is_normalized = Column(Boolean)
    uses_authentic_standard = Column(Boolean)
    authentic_standard = relationship("_ProductMeasurementAuthenticStandard", uselist=False)
    percentage = relationship("_ProductMeasurementPercentage", uselist=False)
    float_value = relationship("_ProductMeasurementFloatValue", uselist=False)
    string_value = Column(Text)
    amount = relationship("_ProductMeasurementAmount", uselist=False)
    retention_time = relationship("_ProductMeasurementRetentionTime", uselist=False)
    mass_spec_details = relationship("MassSpecMeasurementDetails", uselist=False)
    selectivity = relationship("Selectivity", uselist=False)
    wavelength = relationship("_ProductMeasurementWavelength", uselist=False)


class MassSpecMeasurementDetails(Base):
    product_measurement_id = Column(
        Integer, ForeignKey("product_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.ProductMeasurement.MassSpecMeasurementDetails.MassSpecMeasurementType.keys(),
            name="ProductMeasurement.MassSpecMeasurementDetails.MassSpecMeasurementType",
        )
    )
    details = Column(Text)
    tic_minimum_mz = Column(Float)
    tic_maximum_mz = Column(Float)
    eic_masses = relationship("_MassSpecMeasurementDetailsEicMasses")


class Selectivity(Base):
    product_measurement_id = Column(
        Integer, ForeignKey("product_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    type = Column(
        Enum(
            *reaction_pb2.ProductMeasurement.Selectivity.SelectivityType.keys(),
            name="ProductMeasurement.Selectivity.SelectivityType",
        )
    )
    details = Column(Text)


class _DateTime(Parent, Base):
    value = Column(Text)


class _AnalysisInstrumentLastCalibrated(Child, _DateTime):
    analysis_id = Column(Integer, ForeignKey("analysis.id", ondelete="CASCADE"), nullable=False, unique=True)


class _ReactionProvenanceExperimentStart(Child, _DateTime):
    analysis_id = Column(Integer, ForeignKey("reaction_provenance.id", ondelete="CASCADE"), nullable=False, unique=True)


class _RecordEventTime(Child, _DateTime):
    analysis_id = Column(Integer, ForeignKey("record_event.id", ondelete="CASCADE"), nullable=False, unique=True)


class _Analysis(Parent, Base):
    type = Column(Enum(*reaction_pb2.Analysis.AnalysisType.keys(), name="Analysis.AnalysisType"))
    details = Column(Text)
    chmo_id = Column(Integer)
    is_of_isolated_species = Column(Boolean)
    data = relationship("_AnalysisData")
    instrument_manufacturer = Column(Text)
    instrument_last_calibrated = relationship("_AnalysisInstrumentLastCalibrated", uselist=False)


class _CompoundAnalyses(Child, _Analysis):
    compound_id = Column(Integer, ForeignKey("compound.id", ondelete="CASCADE"), nullable=False)
    key = Column(Text)  # Map key.


class _ReactionOutcomeAnalyses(Child, _Analysis):
    reaction_outcome_id = Column(Integer, ForeignKey("reaction_outcome.id", ondelete="CASCADE"), nullable=False)
    key = Column(Text)  # Map key.


class ReactionProvenance(Base):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"), nullable=False, unique=True)

    experimenter = relationship("_ReactionProvenanceExperimenter", uselist=False)
    city = Column(Text)
    experiment_start = relationship("_ReactionProvenanceExperimentStart", uselist=False)
    doi = Column(Text)
    patent = Column(Text)
    publication_url = Column(Text)
    record_created = relationship("_ReactionProvenanceRecordCreated", uselist=False)
    record_modified = relationship("_ReactionProvenanceRecordModified")


class _Person(Parent, Base):
    username = Column(Text)
    name = Column(Text)
    orcid = Column(Text)
    organization = Column(Text)
    email = Column(Text)


class _ReactionProvenanceExperimenter(Child, _Person):
    reaction_provenance_id = Column(
        Integer, ForeignKey("reaction_provenance.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _RecordEventPerson(Child, _Person):
    record_event_id = Column(Integer, ForeignKey("record_event.id", ondelete="CASCADE"), nullable=False, unique=True)


class _RecordEvent(Parent, Base):
    time = relationship("_RecordEventTime", uselist=False)
    person = relationship("_RecordEventPerson", uselist=False)
    details = Column(Text)


class _ReactionProvenanceRecordCreated(Child, _RecordEvent):
    reaction_provenance_id = Column(
        Integer, ForeignKey("reaction_provenance.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ReactionProvenanceRecordModified(Child, _RecordEvent):
    reaction_provenance_id = Column(Integer, ForeignKey("reaction_provenance.id", ondelete="CASCADE"), nullable=False)


class _DatasetExampleCreated(Child, _RecordEvent):
    dataset_example_id = Column(
        Integer, ForeignKey("dataset_example.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _Time(Parent, Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Time.TimeUnit.keys(), name="Time.TimeUnit"))


class _ReactionInputAdditionTime(Child, _Time):
    reaction_input_id = Column(
        Integer, ForeignKey("reaction_input.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ReactionInputAdditionDuration(Child, _Time):
    reaction_input_id = Column(
        Integer, ForeignKey("reaction_input.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _TemperatureConditionsMeasurementTime(Child, _Time):
    temperature_conditions_measurement_id = Column(
        Integer, ForeignKey("temperature_conditions_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _PressureConditionsMeasurementTime(Child, _Time):
    pressure_conditions_measurement_id = Column(
        Integer, ForeignKey("pressure_conditions_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ElectrochemistryConditionsMeasurementTime(Child, _Time):
    electrochemistry_conditions_measurement_id = Column(
        Integer,
        ForeignKey("electrochemistry_conditions_measurement.id", ondelete="CASCADE"),
        nullable=False,
        unique=True,
    )


class _ReactionObservationTime(Child, _Time):
    reaction_observation_id = Column(
        Integer, ForeignKey("reaction_observation.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ReactionWorkupDuration(Child, _Time):
    reaction_workup_id = Column(
        Integer, ForeignKey("reaction_workup.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ReactionOutcomeReactionTime(Child, _Time):
    reaction_outcome_id = Column(
        Integer, ForeignKey("reaction_outcome.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ProductMeasurementRetentionTime(Child, _Time):
    product_measurement_id = Column(
        Integer, ForeignKey("product_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class Mass(Base):
    amount_id = Column(Integer, ForeignKey("amount.id", ondelete="CASCADE"), nullable=False, unique=True)

    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Mass.MassUnit.keys(), name="Mass.MassUnit"))


class Moles(Base):
    amount_id = Column(Integer, ForeignKey("amount.id", ondelete="CASCADE"), nullable=False, unique=True)

    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Moles.MolesUnit.keys(), name="Moles.MolesUnit"))


class _Volume(Parent, Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Volume.VolumeUnit.keys(), name="Volume.VolumeUnit"))


class _AmountVolume(Child, _Volume):
    amount_id = Column(Integer, ForeignKey("amount.id", ondelete="CASCADE"), nullable=False, unique=True)


class _VesselVolume(Child, _Volume):
    vessel_id = Column(Integer, ForeignKey("vessel.id", ondelete="CASCADE"), nullable=False, unique=True)


class Concentration(Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Concentration.ConcentrationUnit.keys(), name="Concentration.ConcentrationUnit"))


class _Pressure(Parent, Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Pressure.PressureUnit.keys(), name="Pressure.PressureUnit"))


class _PressureConditionsSetpoint(Child, _Pressure):
    pressure_conditions_id = Column(
        Integer, ForeignKey("pressure_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _PressureConditionsMeasurementPressure(Child, _Pressure):
    pressure_conditions_measurement_id = Column(
        Integer, ForeignKey("pressure_conditions_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _Temperature(Parent, Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Temperature.TemperatureUnit.keys(), name="Temperature.TemperatureUnit"))


class _ReactionInputAdditionTemperature(Child, _Temperature):
    reaction_input_id = Column(
        Integer, ForeignKey("reaction_input.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _TemperatureConditionsSetpoint(Child, _Temperature):
    temperature_conditions_id = Column(
        Integer, ForeignKey("temperature_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _TemperatureConditionsMeasurementTemperature(Child, _Temperature):
    temperature_conditions_measurement_id = Column(
        Integer, ForeignKey("temperature_conditions_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _Current(Parent, Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Current.CurrentUnit.keys(), name="Current.CurrentUnit"))


class _ElectrochemistryConditionsCurrent(Child, _Current):
    electrochemistry_conditions_id = Column(
        Integer, ForeignKey("electrochemistry_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ElectrochemistryConditionsMeasurementCurrent(Child, _Current):
    electrochemistry_conditions_measurement_id = Column(
        Integer,
        ForeignKey("electrochemistry_conditions_measurement.id", ondelete="CASCADE"),
        nullable=False,
        unique=True,
    )


class _Voltage(Parent, Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Voltage.VoltageUnit.keys(), name="Voltage.VoltageUnit"))


class _ElectrochemistryConditionsVoltage(Child, _Voltage):
    electrochemistry_conditions_id = Column(
        Integer, ForeignKey("electrochemistry_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ElectrochemistryConditionsMeasurementVoltage(Child, _Voltage):
    electrochemistry_conditions_measurement_id = Column(
        Integer,
        ForeignKey("electrochemistry_conditions_measurement.id", ondelete="CASCADE"),
        nullable=False,
        unique=True,
    )


class _Length(Parent, Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Length.LengthUnit.keys(), name="Length.LengthUnit"))


class _IlluminationConditionsDistanceToVessel(Child, _Length):
    illumination_conditions_id = Column(
        Integer, ForeignKey("illumination_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ElectrochemistryConditionsElectrodeSeparation(Child, _Length):
    electrochemistry_conditions_id = Column(
        Integer, ForeignKey("electrochemistry_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _TubingDiameter(Child, _Length):
    tubing_id = Column(Integer, ForeignKey("tubing.id", ondelete="CASCADE"), nullable=False, unique=True)


class _Wavelength(Parent, Base):
    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.Wavelength.WavelengthUnit.keys(), name="Wavelength.WavelengthUnit"))


class _IlluminationConditionsPeakWavelength(Child, _Wavelength):
    illumination_conditions_id = Column(
        Integer, ForeignKey("illumination_conditions.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ProductMeasurementWavelength(Child, _Wavelength):
    product_measurement_id = Column(
        Integer, ForeignKey("product_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class FlowRate(Base):
    reaction_input_id = Column(
        Integer, ForeignKey("reaction_input.id", ondelete="CASCADE"), nullable=False, unique=True
    )

    value = Column(Float)
    precision = Column(Float)
    units = Column(Enum(*reaction_pb2.FlowRate.FlowRateUnit.keys(), name="FlowRate.FlowRateUnit"))


class _Percentage(Parent, Base):
    value = Column(Float)
    precision = Column(Float)


class _ReactionOutcomeConversion(Child, _Percentage):
    reaction_outcome_id = Column(
        Integer, ForeignKey("reaction_outcome.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ProductMeasurementPercentage(Child, _Percentage):
    product_measurement_id = Column(
        Integer, ForeignKey("product_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _FloatValue(Parent, Base):
    value = Column(Float)
    precision = Column(Float)


class _ProductMeasurementFloatValue(Child, _FloatValue):
    product_measurement_id = Column(
        Integer, ForeignKey("product_measurement.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _MassSpecMeasurementDetailsEicMasses(Child, _FloatValue):
    mass_spec_measurement_details_id = Column(
        Integer, ForeignKey("mass_spec_measurement_details.id", ondelete="CASCADE"), nullable=False
    )


class _Data(Parent, Base):
    float_value = Column(Float)
    integer_value = Column(Integer)
    bytes_value = Column(LargeBinary)
    string_value = Column(Text)
    url = Column(Text)
    description = Column(Text)
    format = Column(Text)


class _CompoundFeatures(Child, _Data):
    compound_id = Column(Integer, ForeignKey("compound.id", ondelete="CASCADE"), nullable=False)
    key = Column(Text)  # Map key.


class _ReactionSetupAutomationCode(Child, _Data):
    reaction_setup_id = Column(Integer, ForeignKey("reaction_setup.id", ondelete="CASCADE"), nullable=False)
    key = Column(Text)  # Map key.


class _ReactionObservationImage(Child, _Data):
    reaction_observation_id = Column(
        Integer, ForeignKey("reaction_observation.id", ondelete="CASCADE"), nullable=False, unique=True
    )


class _ProductCompoundFeatures(Child, _Data):
    product_compound_id = Column(Integer, ForeignKey("product_compound.id", ondelete="CASCADE"), nullable=False)
    key = Column(Text)  # Map key.


class _AnalysisData(Child, _Data):
    analysis_id = Column(Integer, ForeignKey("analysis.id", ondelete="CASCADE"), nullable=False)
    key = Column(Text)  # Map key.


_MAPPERS: dict[Type[Message], Type[Base]] = {
    dataset_pb2.Dataset: Dataset,
    dataset_pb2.DatasetExample: DatasetExample,
    reaction_pb2.Reaction: Reaction,
    reaction_pb2.ReactionIdentifier: ReactionIdentifier,
    reaction_pb2.ReactionInput: _ReactionInput,
    reaction_pb2.ReactionInput.AdditionSpeed: AdditionSpeed,
    reaction_pb2.ReactionInput.AdditionDevice: AdditionDevice,
    reaction_pb2.Amount: _Amount,
    reaction_pb2.UnmeasuredAmount: UnmeasuredAmount,
    reaction_pb2.CrudeComponent: CrudeComponent,
    reaction_pb2.Compound: _Compound,
    reaction_pb2.Compound.Source: Source,
    reaction_pb2.CompoundPreparation: CompoundPreparation,
    reaction_pb2.CompoundIdentifier: _CompoundIdentifier,
    reaction_pb2.Vessel: Vessel,
    reaction_pb2.VesselMaterial: VesselMaterial,
    reaction_pb2.VesselAttachment: VesselAttachment,
    reaction_pb2.VesselPreparation: VesselPreparation,
    reaction_pb2.ReactionSetup: ReactionSetup,
    reaction_pb2.ReactionSetup.ReactionEnvironment: ReactionEnvironment,
    reaction_pb2.ReactionConditions: ReactionConditions,
    reaction_pb2.TemperatureConditions: _TemperatureConditions,
    reaction_pb2.TemperatureConditions.TemperatureControl: TemperatureControl,
    reaction_pb2.TemperatureConditions.Measurement: TemperatureConditionsMeasurement,
    reaction_pb2.PressureConditions: PressureConditions,
    reaction_pb2.PressureConditions.PressureControl: PressureControl,
    reaction_pb2.PressureConditions.Atmosphere: Atmosphere,
    reaction_pb2.PressureConditions.Measurement: PressureConditionsMeasurement,
    reaction_pb2.StirringConditions: _StirringConditions,
    reaction_pb2.StirringConditions.StirringRate: StirringRate,
    reaction_pb2.IlluminationConditions: IlluminationConditions,
    reaction_pb2.ElectrochemistryConditions: ElectrochemistryConditions,
    reaction_pb2.ElectrochemistryConditions.Measurement: ElectrochemistryConditionsMeasurement,
    reaction_pb2.ElectrochemistryConditions.ElectrochemistryCell: ElectrochemistryCell,
    reaction_pb2.FlowConditions: FlowConditions,
    reaction_pb2.FlowConditions.Tubing: Tubing,
    reaction_pb2.ReactionNotes: ReactionNotes,
    reaction_pb2.ReactionObservation: ReactionObservation,
    reaction_pb2.ReactionWorkup: ReactionWorkup,
    reaction_pb2.ReactionOutcome: ReactionOutcome,
    reaction_pb2.ProductCompound: ProductCompound,
    reaction_pb2.ProductCompound.Texture: Texture,
    reaction_pb2.ProductMeasurement: ProductMeasurement,
    reaction_pb2.ProductMeasurement.MassSpecMeasurementDetails: MassSpecMeasurementDetails,
    reaction_pb2.ProductMeasurement.Selectivity: Selectivity,
    reaction_pb2.DateTime: _DateTime,
    reaction_pb2.Analysis: _Analysis,
    reaction_pb2.ReactionProvenance: ReactionProvenance,
    reaction_pb2.Person: _Person,
    reaction_pb2.RecordEvent: _RecordEvent,
    reaction_pb2.Time: _Time,
    reaction_pb2.Mass: Mass,
    reaction_pb2.Moles: Moles,
    reaction_pb2.Volume: _Volume,
    reaction_pb2.Concentration: Concentration,
    reaction_pb2.Pressure: _Pressure,
    reaction_pb2.Temperature: _Temperature,
    reaction_pb2.Current: _Current,
    reaction_pb2.Voltage: _Voltage,
    reaction_pb2.Length: _Length,
    reaction_pb2.Wavelength: _Wavelength,
    reaction_pb2.FlowRate: FlowRate,
    reaction_pb2.Percentage: _Percentage,
    reaction_pb2.FloatValue: _FloatValue,
    reaction_pb2.Data: _Data,
}
_PROTOS: dict[Type[Base], Type[Message]] = {value: key for key, value in _MAPPERS.items()}

_MAPPER_RENAMES: dict[tuple[Type[Message], str], str] = {
    (reaction_pb2.Compound.Source, "id"): "vendor_id",
}
_PROTO_RENAMES: dict[tuple[Type[Base], str], str] = {
    (_MAPPERS[key[0]], value): key[1] for key, value in _MAPPER_RENAMES.items()
}


# Polymorphic matchers for constructing queries.
Amount = with_polymorphic(_Amount, "*")
Analysis = with_polymorphic(_Analysis, "*")
Compound = with_polymorphic(_Compound, "*")
CompoundIdentifier = with_polymorphic(_CompoundIdentifier, "*")
Current = with_polymorphic(_Current, "*")
Data = with_polymorphic(_Data, "*")
DateTime = with_polymorphic(_DateTime, "*")
FloatValue = with_polymorphic(_FloatValue, "*")
Length = with_polymorphic(_Length, "*")
Percentage = with_polymorphic(_Percentage, "*")
Person = with_polymorphic(_Person, "*")
Pressure = with_polymorphic(_Pressure, "*")
ReactionInput = with_polymorphic(_ReactionInput, "*")
RecordEvent = with_polymorphic(_RecordEvent, "*")
StirringConditions = with_polymorphic(_StirringConditions, "*")
Temperature = with_polymorphic(_Temperature, "*")
TemperatureConditions = with_polymorphic(_TemperatureConditions, "*")
Time = with_polymorphic(_Time, "*")
Voltage = with_polymorphic(_Voltage, "*")
Volume = with_polymorphic(_Volume, "*")
Wavelength = with_polymorphic(_Wavelength, "*")


def from_proto(  # pylint: disable=too-many-branches
    message: Message, mapper: Optional[Type[Base]] = None, key: Optional[str] = None
) -> Base:
    """Converts a protobuf message into an ORM object.

    Args:
        message: Protobuf message.
        mapper: ORM mapper class. For top-level protos like Dataset and Reaction this can be left as None; it must
            be provided for Child subclasses to properly handle polymorphism.
        key: Map key (we store maps as rows of (key, value) tuples).

    Returns:
        ORM object.
    """
    if mapper is None:
        mapper = _MAPPERS[type(message)]
    kwargs = {}
    if key is not None:
        kwargs["key"] = key
    if mapper == Reaction:
        kwargs["proto"] = message.SerializeToString()
    for field, value in message.ListFields():
        field_name = _MAPPER_RENAMES.get((type(message), field.name), field.name)
        if field_name == "eic_masses":
            # Convert repeated float to repeated FloatValue.
            kwargs[field_name] = [_MassSpecMeasurementDetailsEicMasses(value=v) for v in value]
        elif field_name == "reaction_ids":
            # Convert repeated string to repeated ReactionId.
            kwargs[field_name] = [ReactionId(reaction_id=v) for v in value]
        elif field.type == FieldDescriptor.TYPE_MESSAGE:
            field_mapper = getattr(mapper, field_name).mapper.class_
            if isinstance(value, Mapping):
                kwargs[field_name] = [from_proto(v, mapper=field_mapper, key=k) for k, v in value.items()]
            elif field.label == FieldDescriptor.LABEL_REPEATED:
                kwargs[field_name] = [from_proto(v, mapper=field_mapper) for v in value]
            else:
                kwargs[field_name] = from_proto(value, mapper=field_mapper)
        elif field.type == FieldDescriptor.TYPE_ENUM:
            kwargs[field_name] = field.enum_type.values_by_number[value].name
        else:
            kwargs[field_name] = value
    if isinstance(message, (reaction_pb2.Compound, reaction_pb2.ProductCompound)):
        # Add RDKit cartridge functionality.
        field_mapper = getattr(mapper, "structure").mapper.class_
        try:
            kwargs["structure"] = field_mapper(smiles=message_helpers.smiles_from_compound(message))
        except ValueError:
            pass
    return mapper(**kwargs)


def to_proto(base: Base) -> Message:
    """Converts an ORM object into a protobuf message.

    Args:
        base: ORM object.

    Returns:
        Protobuf message.
    """
    kwargs = {}
    proto = None
    for mapper in getmro(type(base)):
        if not issubclass(mapper, Child):
            proto = _PROTOS[mapper]
            break
    assert issubclass(proto, Message)
    for field in proto.DESCRIPTOR.fields:
        mapper_field_name = _MAPPER_RENAMES.get((proto, field.name), field.name)
        value = getattr(base, mapper_field_name)
        field_name = _PROTO_RENAMES.get((type(base), mapper_field_name), mapper_field_name)
        if field_name == "eic_masses":
            # Convert repeated FloatValue to repeated float.
            kwargs[field_name] = [v.value for v in value]
        elif field_name == "reaction_ids":
            # Convert repeated ReactionId to repeated string.
            kwargs[field_name] = [v.reaction_id for v in value]
        elif isinstance(value, list):
            if len(value) == 0:
                continue
            if hasattr(value[0], "key"):
                kwargs[field_name] = {v.key: to_proto(v) for v in value}
            else:
                kwargs[field_name] = [to_proto(v) for v in value]
        elif isinstance(value, Base):
            kwargs[field_name] = to_proto(value)
        else:
            kwargs[field_name] = value
    return proto(**kwargs)



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers validating specific Message types."""
import dataclasses
import math
import os
import re
import warnings
from collections.abc import Mapping
from typing import Any, Optional

from dateutil import parser
from rdkit import Chem
from rdkit import __version__ as RDKIT_VERSION

import ord_schema
from ord_schema.logging import get_logger
from ord_schema import message_helpers
from ord_schema.proto import dataset_pb2
from ord_schema.proto import reaction_pb2

logger = get_logger(__name__)


# pylint: disable=too-many-branches


@dataclasses.dataclass
class ValidationOptions:
    """Options for message validation."""

    # Check that Dataset and Reaction IDs are well-formed.
    validate_ids: bool = False
    # Require ReactionProvenance for Reactions.
    require_provenance: bool = False
    # Allow reactions with valid reaction SMILES and nothing else.
    allow_reaction_smiles_only: bool = True


@dataclasses.dataclass
class ValidationOutput:
    """Validation output: errors and warnings."""

    errors: list[str] = dataclasses.field(default_factory=list)
    warnings: list[str] = dataclasses.field(default_factory=list)

    def extend(self, other):
        self.errors.extend(other.errors)
        self.warnings.extend(other.warnings)


def validate_datasets(
    datasets: Mapping[str, dataset_pb2.Dataset],
    write_errors: bool = False,
    options: Optional[ValidationOptions] = None,
) -> None:
    """Runs validation for a set of datasets.

    Args:
        datasets: Dict mapping text filenames to Dataset protos.
        write_errors: If True, errors are written to disk.
        options: ValidationOptions.

    Raises:
        ValidationError: if any Dataset does not pass validation.
    """
    all_errors = []
    for filename, dataset in datasets.items():
        basename = os.path.basename(filename)
        errors = _validate_datasets(dataset, label=basename, options=options)
        if errors:
            for error in errors:
                all_errors.append(f"{filename}: {error}")
            if write_errors:
                with open(f"{filename}.error", "w") as f:
                    for error in errors:
                        f.write(f"{error}\n")
    # NOTE(kearnes): We run validation for all datasets before exiting if there
    # are errors.
    if all_errors:
        error_string = "\n".join(all_errors)
        raise ValidationError(f"validation encountered errors:\n{error_string}")


def _validate_datasets(
    dataset: dataset_pb2.Dataset,
    label: str = "dataset",
    options: Optional[ValidationOptions] = None,
) -> list[str]:
    """Validates Reaction messages and cross-references in a Dataset.

    Args:
        dataset: dataset_pb2.Dataset message.
        label: string label for logging purposes only.
        options: ValidationOptions.

    Returns:
        List of validation error messages.
    """
    errors = []
    # Reaction-level validation.
    num_bad_reactions = 0
    for i, reaction in enumerate(dataset.reactions):
        reaction_output = validate_message(reaction, raise_on_error=False, options=options)
        if reaction_output.errors:
            num_bad_reactions += 1
        for error in reaction_output.errors:
            errors.append(error)
            logger.warning(f"Validation error for {label}[{i}]: {error}")
    num_successful = (len(dataset.reactions) - num_bad_reactions,)
    logger.info(
        f"Validation summary for {label}: {num_successful}/{len(dataset.reactions)} successful "
        f"({num_bad_reactions} failures)"
    )
    # Dataset-level validation of cross-references.
    dataset_output = validate_message(dataset, raise_on_error=False, recurse=False, options=options)
    for error in dataset_output.errors:
        errors.append(error)
        logger.warning(f"Validation error for {label}: {error}")

    return errors


def validate_message(
    message: ord_schema.Message,
    recurse: bool = True,
    raise_on_error: bool = True,
    options: Optional[ValidationOptions] = None,
    trace: Optional[tuple[str, ...]] = None,
) -> ValidationOutput:
    """Template function for validating custom messages in the reaction_pb2.

    Messages are not validated to check enum values, since these are enforced
    by the schema. Instead, we only check for validity of items that cannot be
    enforced in the schema (e.g., non-negativity of certain measurements,
    consistency of cross-referenced keys).

    Note that the message may be modified in-place with any unambiguous changes
    needed to ensure validity.

    Args:
        message: A message to validate.
        recurse: A boolean that controls whether submessages of message (i.e.,
            fields that are messages) should also be validated. Defaults to
            True.
        raise_on_error: If True, raises a ValidationError exception when errors
            are encountered. If False, the user must manually check the return
            value to identify validation errors.
        options: ValidationOptions.
        trace: Tuple containing a string "stack trace" to track the position of
            the current message relative to the recursion root.

    Returns:
        ValidationOutput.

    Raises:
        ValidationError: If any fields are invalid.
    """
    if trace is None:
        trace = (message.DESCRIPTOR.name,)
    output = ValidationOutput()
    # Recurse through submessages
    if recurse:
        for field, value in message.ListFields():
            if field.type == field.TYPE_MESSAGE:  # need to recurse
                _validate_message(
                    field=field,
                    value=value,
                    output=output,
                    raise_on_error=raise_on_error,
                    options=options,
                    trace=trace,
                )

    # Message-specific validation
    if not isinstance(message, tuple(_VALIDATOR_SWITCH.keys())):
        # NOTE(ccoley): I made the conscious decision to raise an error here,
        # rather than assume that the message is valid. If a message does not
        # require any message-level checks (not uncommon), then it should still
        # be listed in the dictionary switch above withpass. This will force
        # us to think about what is necessary if/when new messages are added.
        raise NotImplementedError(f"Don't know how to validate {type(message)}")

    with warnings.catch_warnings(record=True) as tape:
        if isinstance(message, (reaction_pb2.Reaction, dataset_pb2.Dataset)):
            _VALIDATOR_SWITCH[type(message)](message, options=options)
        else:
            _VALIDATOR_SWITCH[type(message)](message)
    stack = ".".join(trace)
    for warning in tape:
        message = f"{stack}: {warning.message}"
        if issubclass(warning.category, ValidationError):
            if raise_on_error:
                raise ValidationError(message)
            output.errors.append(message)
        else:
            output.warnings.append(message)
    return output


def _validate_message(
    field: ord_schema.FieldDescriptor,
    value: Any,
    output: ValidationOutput,
    raise_on_error: bool,
    options: ValidationOptions,
    trace: tuple[str, ...],
):
    """Validates a single message field and its children.

    Args:
        field: FieldDescriptor instance.
        value: The value of the current message field.
        output: ValidationOutput.
        raise_on_error: If True, raises a ValidationError exception when errors
            are encountered. If False, the user must manually check the return
            value to identify validation errors.
        options: ValidationOptions.
        trace: Tuple containing a string "stack trace" to track the position of
            the current message relative to the recursion root.
    """
    if field.label == field.LABEL_REPEATED:
        if field.message_type.GetOptions().map_entry:  # map
            # value is message
            if field.message_type.fields_by_name["value"].type == field.TYPE_MESSAGE:
                for key, submessage in value.items():
                    this_trace = trace + (f'{field.name}["{key}"]',)
                    this_output = validate_message(
                        submessage,
                        raise_on_error=raise_on_error,
                        options=options,
                        trace=this_trace,
                    )
                    output.extend(this_output)
            else:  # value is a primitive
                pass
        else:  # Just a repeated message
            for index, submessage in enumerate(value):
                this_trace = trace + (f"{field.name}[{index}]",)
                this_output = validate_message(
                    submessage,
                    raise_on_error=raise_on_error,
                    options=options,
                    trace=this_trace,
                )
                output.extend(this_output)
    else:  # no recursion needed
        this_trace = trace + (field.name,)
        this_output = validate_message(value, raise_on_error=raise_on_error, options=options, trace=this_trace)
        output.extend(this_output)


class ValidationError(Warning):
    pass


class ValidationWarning(Warning):
    pass


def is_empty(message: ord_schema.Message):
    """Returns whether the given message is empty."""
    empty = type(message)().SerializeToString()
    return message.SerializeToString(deterministic=True) == empty


# pylint: disable=missing-function-docstring
def ensure_float_nonnegative(message: ord_schema.Message, field: str):
    if getattr(message, field) < 0:
        warnings.warn(
            f"Field {field} of message " f"{type(message).DESCRIPTOR.name} must be non-negative",
            ValidationError,
        )


def ensure_float_range(
    message: ord_schema.Message,
    field: str,
    min_value: float = -math.inf,
    max_value: float = math.inf,
):
    if getattr(message, field) < min_value or getattr(message, field) > max_value:
        warnings.warn(
            f"Field {field} of message "
            f"{type(message).DESCRIPTOR.name} must be between"
            f" {min_value} and {max_value}",
            ValidationError,
        )


def check_value_and_units(message: ord_schema.UnitMessage):
    """Checks that value/units messages are complete."""
    if not message.HasField("value"):
        warnings.warn(f"{type(message)} requires `value` to be set", ValidationError)
    if message.units == message.UNSPECIFIED:
        warnings.warn(f"{type(message)} requires `units` to be set", ValidationError)


def check_type_and_details(message: ord_schema.TypeDetailsMessage):
    """Checks that type/details messages are complete."""
    if is_empty(message):
        return
    if message.type == message.UNSPECIFIED:
        warnings.warn(f"{type(message)} requires `type` to be set", ValidationError)
    if message.type == message.CUSTOM and not message.details:
        warnings.warn(
            f"{type(message)} has type CUSTOM but details field is empty",
            ValidationError,
        )


def reaction_has_internal_standard(message: reaction_pb2.Reaction) -> bool:
    """Whether any reaction component uses the internal standard role."""
    for reaction_input in message.inputs.values():
        for compound in reaction_input.components:
            if compound.reaction_role == reaction_pb2.ReactionRole.INTERNAL_STANDARD:
                return True
    for workup in message.workups:
        if workup.input:
            for compound in workup.input.components:
                if compound.reaction_role == reaction_pb2.ReactionRole.INTERNAL_STANDARD:
                    return True
    return False


def reaction_has_limiting_component(message: reaction_pb2.Reaction) -> bool:
    """Whether any reaction input compound is limiting."""
    for reaction_input in message.inputs.values():
        for compound in reaction_input.components:
            if compound.is_limiting:
                return True
    return False


def reaction_needs_internal_standard(message: reaction_pb2.Reaction) -> bool:
    """Whether any analysis uses an internal standard."""
    for outcome in message.outcomes:
        for product in outcome.products:
            for measurement in product.measurements:
                if measurement.uses_internal_standard:
                    return True
    return False


def get_referenced_reaction_ids(message: reaction_pb2.Reaction) -> set[str]:
    """Return the set of reaction IDs that are referenced in a Reaction."""
    referenced_ids = set()
    for reaction_input in message.inputs.values():
        for component in reaction_input.components:
            for preparation in component.preparations:
                if preparation.reaction_id:
                    referenced_ids.add(preparation.reaction_id)
        for crude_component in reaction_input.crude_components:
            referenced_ids.add(crude_component.reaction_id)
    return referenced_ids


def is_valid_reaction_id(reaction_id: str) -> bool:
    match = re.fullmatch("^ord-[0-9a-f]{32}$", reaction_id)
    return bool(match)


def is_valid_dataset_id(dataset_id: str) -> bool:
    match = re.fullmatch("^ord_dataset-[0-9a-f]{32}$", dataset_id)
    return bool(match)


def validate_dataset(message: dataset_pb2.Dataset, options: Optional[ValidationOptions] = None):
    # pylint: disable=too-many-branches,too-many-nested-blocks
    if options is None:
        options = ValidationOptions()
    if not message.reactions and not message.reaction_ids:
        warnings.warn("Dataset requires reactions or reaction_ids", ValidationError)
    elif message.reactions and message.reaction_ids:
        warnings.warn("Dataset requires reactions or reaction_ids, not both", ValidationError)
    if message.reaction_ids:
        for reaction_id in message.reaction_ids:
            if not is_valid_reaction_id(reaction_id):
                warnings.warn("Reaction ID is malformed", ValidationError)
    if options.validate_ids:
        # The dataset_id is a 32-character uuid4 hex string.
        if not is_valid_dataset_id(message.dataset_id):
            warnings.warn("Dataset ID is malformed", ValidationError)
    # Check cross-references
    dataset_referenced_ids = set()
    dataset_defined_ids = set()
    for reaction in message.reactions:
        if reaction.reaction_id:
            if reaction.reaction_id in dataset_defined_ids:
                warnings.warn(
                    "Multiple Reactions should never have the same IDs",
                    ValidationError,
                )
            dataset_defined_ids.add(reaction.reaction_id)
        referenced_ids = get_referenced_reaction_ids(reaction)
        if any(_id == reaction.reaction_id for _id in referenced_ids):
            warnings.warn("A Reaction should not reference its own ID", ValidationError)
        dataset_referenced_ids |= referenced_ids
    if len(dataset_referenced_ids - dataset_defined_ids) > 0:
        warnings.warn(
            "Reactions in the Dataset refer to undefined "
            f"reaction_ids {dataset_referenced_ids - dataset_defined_ids}",
            ValidationError,
        )


def validate_dataset_example(message: dataset_pb2.DatasetExample):
    if not message.description:
        warnings.warn("DatasetExample.description is required", ValidationError)
    if not message.url:
        warnings.warn("DatasetExample.url is required", ValidationError)
    if not message.HasField("created"):
        warnings.warn("DatasetExample.created is required", ValidationError)


def validate_reaction(message: reaction_pb2.Reaction, options: Optional[ValidationOptions] = None):
    if options is None:
        options = ValidationOptions()
    if (
        options.allow_reaction_smiles_only
        and message_helpers.get_reaction_smiles(message)
        and len(message.inputs) == 0
        and len(message.outcomes) == 0
    ):
        pass
    else:
        if len(message.inputs) == 0:
            warnings.warn("Reactions should have at least 1 reaction input", ValidationError)
        if len(message.outcomes) == 0:
            warnings.warn("Reactions should have at least 1 reaction outcome", ValidationError)
    for input_ in message.inputs:
        for component in message.inputs[input_].components:
            if not component.amount.WhichOneof("kind"):
                warnings.warn("All reaction input components require an amount", ValidationError)
    if reaction_needs_internal_standard(message) and not reaction_has_internal_standard(message):
        warnings.warn(
            "Reaction analysis uses an internal standard, but no "
            "component (as reaction input or workup) uses the "
            "reaction role INTERNAL_STANDARD",
            ValidationError,
        )
    if any(outcome.HasField("conversion") for outcome in message.outcomes) and not reaction_has_limiting_component(
        message
    ):
        warnings.warn(
            "If reaction conversion is specified, at least one reaction input component must be labeled is_limiting",
            ValidationError,
        )
    if options.validate_ids:
        # The reaction_id suffix is a 32-character uuid4 hex string.
        if not re.fullmatch("^ord-[0-9a-f]{32}$", message.reaction_id):
            warnings.warn("Reaction ID is malformed", ValidationError)
    if options.require_provenance:
        if not message.HasField("provenance"):
            warnings.warn("Reaction requires provenance", ValidationError)


def validate_reaction_identifier(message: reaction_pb2.ReactionIdentifier):
    check_type_and_details(message)
    if message.type in [message.REACTION_SMILES, message.REACTION_CXSMILES]:
        if message.type == message.REACTION_CXSMILES:
            smiles = message.value.split()[0]
        else:
            smiles = message.value
        try:
            message_helpers.validate_reaction_smiles(smiles)
        except ValueError as error:
            warnings.warn(str(error), ValidationError)
    if not message.value:
        warnings.warn("value must be set", ValidationError)


def validate_reaction_input(message: reaction_pb2.ReactionInput):
    if len(message.components) + len(message.crude_components) == 0:
        warnings.warn("Reaction inputs must have at least one component", ValidationError)
    elif len(message.components) + len(message.crude_components) == 1:
        for component in message.components:
            if (
                component.amount.WhichOneof("kind") == "unmeasured"
                and component.amount.unmeasured.type == reaction_pb2.UnmeasuredAmount.SATURATED
            ):
                warnings.warn(
                    "SATURATED compound amounts should only be used "
                    "for solutes when another component (solvent) is present",
                    ValidationWarning,
                )


def validate_addition_device(message: reaction_pb2.ReactionInput.AdditionDevice):
    check_type_and_details(message)


def validate_addition_speed(message: reaction_pb2.ReactionInput.AdditionSpeed):
    del message  # Unused.


def validate_amount(message: reaction_pb2.Amount):
    if message.HasField("volume_includes_solutes") and message.WhichOneof("kind") != "volume":
        warnings.warn(
            "volume_includes_solutes should only be set for volume amounts",
            ValidationError,
        )


def validate_unmeasured_amount(message: reaction_pb2.UnmeasuredAmount):
    check_type_and_details(message)


def validate_source(message: reaction_pb2.Compound.Source):
    del message  # Unused.


def validate_crude_component(message: reaction_pb2.CrudeComponent):
    if not message.reaction_id:
        warnings.warn("CrudeComponents must specify a reaction_id", ValidationError)
    if message.has_derived_amount and message.amount.HasField("kind"):
        warnings.warn(
            "CrudeComponents with derived amounts cannot have their mass or volume specified explicitly",
            ValidationError,
        )
    if (not message.HasField("has_derived_amount") or not message.has_derived_amount) and not message.amount.HasField(
        "kind"
    ):
        warnings.warn(
            "Crude components should either have a derived amount or a specified mass or volume",
            ValidationError,
        )
    if message.amount.WhichOneof("kind") not in [None, "mass", "volume"]:
        warnings.warn(
            "Crude component amounts must be specified by mass or volume",
            ValidationError,
        )
    if message.amount.HasField("volume_includes_solutes"):
        warnings.warn(
            "volume_includes_solutes should only be used for input Compounds",
            ValidationError,
        )


def validate_compound(message: reaction_pb2.Compound):
    if len(message.identifiers) == 0:
        warnings.warn("Compounds must have at least one identifier", ValidationError)
    if all(identifier.type == identifier.NAME for identifier in message.identifiers):
        warnings.warn(
            "Compounds should have more specific identifiers than NAME whenever possible",
            ValidationWarning,
        )
    try:
        message_helpers.check_compound_identifiers(message)
    except ValueError as error:
        warnings.warn(str(error), ValidationWarning)


def validate_compound_preparation(message: reaction_pb2.CompoundPreparation):
    check_type_and_details(message)
    if message.reaction_id and message.type != message.SYNTHESIZED:
        warnings.warn(
            "Reaction IDs should only be specified in compound preparations when SYNTHESIZED",
            ValidationError,
        )


def validate_compound_identifier(message: reaction_pb2.CompoundIdentifier):
    check_type_and_details(message)
    if not message.value:
        warnings.warn("value must be set", ValidationError)
    if message.type == message.SMILES:
        mol = Chem.MolFromSmiles(message.value)
        if mol is None:
            warnings.warn(
                f"RDKit {RDKIT_VERSION} could not validate" f" SMILES identifier {message.value}",
                ValidationError,
            )
    elif message.type == message.INCHI:
        mol = Chem.MolFromInchi(message.value)
        if mol is None:
            warnings.warn(
                f"RDKit {RDKIT_VERSION} could not validate" f" InChI identifier {message.value}",
                ValidationError,
            )
    elif message.type == message.MOLBLOCK:
        mol = Chem.MolFromMolBlock(message.value)
        if mol is None:
            warnings.warn(
                f"RDKit {RDKIT_VERSION} could not validate MolBlock identifier",
                ValidationError,
            )


def validate_vessel(message: reaction_pb2.Vessel):
    check_type_and_details(message)


def validate_vessel_material(message: reaction_pb2.VesselMaterial):
    check_type_and_details(message)


def validate_vessel_attachment(message: reaction_pb2.VesselAttachment):
    check_type_and_details(message)


def validate_vessel_preparation(message: reaction_pb2.VesselPreparation):
    check_type_and_details(message)


def validate_reaction_setup(message: reaction_pb2.ReactionSetup):
    del message  # Unused.


def validate_reaction_environment(
    message: reaction_pb2.ReactionSetup.ReactionEnvironment,
):
    check_type_and_details(message)


def validate_reaction_conditions(message: reaction_pb2.ReactionConditions):
    if message.conditions_are_dynamic and not message.details:
        warnings.warn(
            "Reaction conditions are dynamic, but no details"
            " provided to explain how procedure deviates from"
            " normal single-step reaction conditions.",
            ValidationError,
        )
    if message.details and not message.conditions_are_dynamic:
        warnings.warn(
            "Reaction condition details provided but field "
            "conditions_are_dynamic is False. If the conditions "
            "cannot be fully captured by the schema, set to True.",
            ValidationWarning,
        )


def validate_temperature_conditions(message: reaction_pb2.TemperatureConditions):
    del message  # Unused.


def validate_temperature_control(
    message: reaction_pb2.TemperatureConditions.TemperatureControl,
):
    check_type_and_details(message)


def validate_temperature_measurement(
    message: reaction_pb2.TemperatureConditions.Measurement,
):
    check_type_and_details(message)


def validate_pressure_conditions(message: reaction_pb2.PressureConditions):
    del message


def validate_pressure_control(message: reaction_pb2.PressureConditions.PressureControl):
    check_type_and_details(message)


def validate_atmosphere(message: reaction_pb2.PressureConditions.Atmosphere):
    check_type_and_details(message)


def validate_pressure_measurement(message: reaction_pb2.PressureConditions.Measurement):
    check_type_and_details(message)


def validate_stirring_conditions(message: reaction_pb2.StirringConditions):
    check_type_and_details(message)


def validate_stirring_rate(message: reaction_pb2.StirringConditions.StirringRate):
    ensure_float_nonnegative(message, "rpm")


def validate_illumination_conditions(message: reaction_pb2.IlluminationConditions):
    check_type_and_details(message)


def validate_electrochemistry_conditions(
    message: reaction_pb2.ElectrochemistryConditions,
):
    check_type_and_details(message)


def validate_electrochemistry_cell(
    message: reaction_pb2.ElectrochemistryConditions.ElectrochemistryCell,
):
    check_type_and_details(message)


def validate_electrochemistry_measurement(
    message: reaction_pb2.ElectrochemistryConditions.Measurement,
):
    del message  # Unused.


def validate_flow_conditions(message: reaction_pb2.FlowConditions):
    check_type_and_details(message)


def validate_tubing(message: reaction_pb2.FlowConditions.Tubing):
    check_type_and_details(message)


def validate_reaction_notes(message: reaction_pb2.ReactionNotes):
    del message  # Unused.


def validate_reaction_observation(message: reaction_pb2.ReactionObservation):
    del message  # Unused.


def validate_reaction_workup(message: reaction_pb2.ReactionWorkup):
    check_type_and_details(message)
    if message.type == reaction_pb2.ReactionWorkup.WAIT and not message.duration.value:
        warnings.warn("WAIT workup steps should have a defined duration", ValidationWarning)
    if message.type == reaction_pb2.ReactionWorkup.TEMPERATURE and not message.HasField("temperature"):
        warnings.warn(
            "TEMPERATURE workup steps should have defined temperature conditions",
            ValidationWarning,
        )
    if (
        message.type
        in (
            reaction_pb2.ReactionWorkup.EXTRACTION,
            reaction_pb2.ReactionWorkup.FILTRATION,
        )
        and not message.keep_phase
    ):
        warnings.warn(
            "Workup step EXTRACTION or FILTRATION missing a recommended field keep_phase",
            ValidationWarning,
        )
    if (
        message.type
        in (
            reaction_pb2.ReactionWorkup.ADDITION,
            reaction_pb2.ReactionWorkup.WASH,
            reaction_pb2.ReactionWorkup.DRY_WITH_MATERIAL,
            reaction_pb2.ReactionWorkup.SCAVENGING,
            reaction_pb2.ReactionWorkup.DISSOLUTION,
            reaction_pb2.ReactionWorkup.PH_ADJUST,
        )
        and not message.input.components
    ):
        warnings.warn("Workup step missing recommended inputs definition", ValidationWarning)
    if message.type == reaction_pb2.ReactionWorkup.STIRRING and not message.stirring:
        warnings.warn("Stirring workup step missing stirring definition", ValidationWarning)
    if message.type == reaction_pb2.ReactionWorkup.PH_ADJUST and not message.HasField("target_ph"):
        warnings.warn("pH adjustment workup missing target pH", ValidationWarning)
    if message.type == reaction_pb2.ReactionWorkup.ALIQUOT:
        if message.amount.WhichOneof("kind") is None:
            warnings.warn("Aliquot workup step missing volume/mass amount", ValidationWarning)
        elif message.amount.WhichOneof("kind") not in ["mass", "volume"]:
            warnings.warn(
                "Aliquot amounts should be specified by mass or volume",
                ValidationWarning,
            )
        if message.amount.HasField("volume_includes_solutes"):
            warnings.warn(
                "volume_includes_solutes should only be used for input Compounds",
                ValidationWarning,
            )
    # Question: Are there other reaction workup types with specifiable amounts?
    if message.amount.WhichOneof("kind") is not None and message.type not in (
        reaction_pb2.ReactionWorkup.ALIQUOT,
        reaction_pb2.ReactionWorkup.CUSTOM,
    ):
        warnings.warn(
            "Workup amount should only be specified if workup type is ALIQUOT or CUSTOM",
            ValidationWarning,
        )


def validate_reaction_outcome(message: reaction_pb2.ReactionOutcome):
    # pylint: disable=singleton-comparison
    # Can only have one desired product
    if sum(product.is_desired_product for product in message.products) > 1:
        warnings.warn("Cannot have more than one desired product!", ValidationError)
    # Check key values for product analyses
    # NOTE(ccoley): Could use any(), but using expanded loops for clarity
    analysis_keys = list(message.analyses.keys())
    for product in message.products:
        for measurement in product.measurements:
            if measurement.analysis_key and measurement.analysis_key not in analysis_keys:
                warnings.warn(
                    f"analysis key {measurement.analysis_key} does not match " f"any known analysis ({analysis_keys})",
                    ValidationError,
                )
    # TODO(ccoley): While we do not currently check whether the parent Reaction
    # is *actually* used in a multistep reaction within a Dataset (i.e., in a
    # CrudeComponent); this is an additional check that could be added to the
    # submission pipeline.
    if not message.products and not message.HasField("conversion"):
        warnings.warn(
            "No products or conversion are specified for reaction; this is permissible only for multistep reactions",
            ValidationWarning,
        )


def validate_product_compound(message: reaction_pb2.ProductCompound):
    if len(message.identifiers) == 0:
        warnings.warn("Compounds must have at least one identifier", ValidationError)
    if all(identifier.type == identifier.NAME for identifier in message.identifiers):
        warnings.warn(
            "Compounds should have more specific identifiers than NAME whenever possible",
            ValidationWarning,
        )
    try:
        message_helpers.check_compound_identifiers(message)
    except ValueError as error:
        warnings.warn(str(error), ValidationWarning)


def validate_texture(message: reaction_pb2.ProductCompound.Texture):
    check_type_and_details(message)


def validate_product_measurement(message: reaction_pb2.ProductMeasurement):
    check_type_and_details(message)
    if not message.analysis_key:
        warnings.warn(
            "Product measurements should be associated with an analysis through its analysis_key",
            ValidationWarning,
        )
    if message.type == reaction_pb2.ProductMeasurement.IDENTITY:
        if message.WhichOneof("value"):
            warnings.warn(
                "Product measurements to confirm IDENTITY should not have any values defined",
                ValidationError,
            )
    elif message.type == reaction_pb2.ProductMeasurement.YIELD:
        if message.WhichOneof("value") != "percentage":
            warnings.warn(
                "YIELD measurements should be defined as percentage values if possible",
                ValidationWarning,
            )
    elif message.type == reaction_pb2.ProductMeasurement.PURITY:
        if message.WhichOneof("value") != "percentage":
            warnings.warn(
                "PURITY measurements should be defined as percentage values if possible",
                ValidationWarning,
            )
    elif message.type in (
        reaction_pb2.ProductMeasurement.AREA,
        reaction_pb2.ProductMeasurement.COUNTS,
        reaction_pb2.ProductMeasurement.INTENSITY,
    ):
        if message.WhichOneof("value") not in ("percentage", "float_value"):
            warnings.warn(
                "Product measurements of type AREA, COUNTS, or "
                "INTENSITY must use numeric values (percentage or float_value)",
                ValidationError,
            )
    if message.HasField("selectivity") and (message.type != reaction_pb2.ProductMeasurement.SELECTIVITY):
        warnings.warn(
            "The selectivity_type field should only be used for a product measurement with type SELECTIVITY",
            ValidationError,
        )


def validate_selectivity(message: reaction_pb2.ProductMeasurement.Selectivity):
    check_type_and_details(message)


def validate_mass_spec_measurement_type(
    message: reaction_pb2.ProductMeasurement.MassSpecMeasurementDetails,
):
    check_type_and_details(message)


def validate_date_time(message: reaction_pb2.DateTime):
    if message.value:
        try:
            parser.parse(message.value).ctime()
        except parser.ParserError:
            warnings.warn(f"Could not parse DateTime string {message.value}", ValidationError)


def validate_analysis(message: reaction_pb2.Analysis):
    # TODO(ccoley): Will be lots to expand here if we add structured data.
    check_type_and_details(message)


def validate_reaction_provenance(message: reaction_pb2.ReactionProvenance):
    # Prepare datetimes
    if not message.HasField("record_created"):
        warnings.warn("Reactions must have record_created defined", ValidationError)
    experiment_start = None
    record_created = None
    record_modified = None
    try:
        if message.experiment_start.value:
            experiment_start = parser.parse(message.experiment_start.value)
        if message.record_created.time.value:
            record_created = parser.parse(message.record_created.time.value)
        for record in message.record_modified:
            # Use the last record as the most recent modification time.
            record_modified = parser.parse(record.time.value)
    except parser.ParserError:
        warnings.warn("Failed to parse DateTime string(s)")
    # Check signs of time differences
    if experiment_start and record_created:
        if (record_created - experiment_start).total_seconds() < 0:
            warnings.warn("Record creation time should be after experiment", ValidationError)
    if record_modified and record_created:
        if (record_modified - record_created).total_seconds() < 0:
            warnings.warn("Record modified time should be after creation", ValidationError)
    if not message.record_created.person.email:
        warnings.warn("User email is required for record_created", ValidationError)
    for record in message.record_modified:
        if not record.person.email:
            warnings.warn("User email is required for record_modified", ValidationError)
    if message.doi:
        parsed_doi = message_helpers.parse_doi(message.doi)
        if message.doi != parsed_doi:
            warnings.warn(
                f"DOI should be trimmed ({message.doi} -> {parsed_doi})",
                ValidationError,
            )
    # TODO(ccoley) could check if publication_url is valid, etc.


def validate_record_event(message: reaction_pb2.RecordEvent):
    if not message.time.value:
        warnings.warn("RecordEvent must have `time` specified", ValidationError)
    person = message.person
    if not (person.username or person.name or person.orcid):
        warnings.warn(
            "Person must have at least one of `username`, `name`, or `orcid` specified",
            ValidationError,
        )
    if not person.email:
        warnings.warn("Person must have `email` specified", ValidationError)


def validate_person(message: reaction_pb2.Person):
    # NOTE(ccoley): final character is checksum, but ignoring that for now
    if message.orcid:
        if not re.match("[0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{3}[0-9X]", message.orcid):
            warnings.warn("Invalid ORCID: Enter as 0000-0000-0000-0000", ValidationError)
    if message.email:
        # Based on https://www.regular-expressions.info/email.html.
        # Added optional "[bot]" suffix to the username for GitHub actions.
        if not re.fullmatch(r"[a-zA-Z0-9._+-]+(?:\[bot\])?@[a-zA-Z0-9.-]+\.[a-z]{2,}", message.email):
            warnings.warn(f"Invalid email address: {message.email}", ValidationError)


def validate_time(message: reaction_pb2.Time):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_mass(message: reaction_pb2.Mass):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_moles(message: reaction_pb2.Moles):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_volume(message: reaction_pb2.Volume):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_concentration(message: reaction_pb2.Concentration):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_pressure(message: reaction_pb2.Pressure):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_temperature(message: reaction_pb2.Temperature):
    check_value_and_units(message)
    if message.units == message.CELSIUS:
        ensure_float_range(message, "value", min_value=-273.15)
    elif message.units == message.FAHRENHEIT:
        ensure_float_range(message, "value", min_value=-459)
    elif message.units == message.KELVIN:
        ensure_float_range(message, "value", min_value=0)
    ensure_float_nonnegative(message, "precision")


def validate_current(message: reaction_pb2.Current):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_voltage(message: reaction_pb2.Voltage):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_length(message: reaction_pb2.Length):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_wavelength(message: reaction_pb2.Wavelength):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_flow_rate(message: reaction_pb2.FlowRate):
    check_value_and_units(message)
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_percentage(message: reaction_pb2.Percentage):
    if not message.HasField("value"):
        warnings.warn(f"{type(message)} requires `value` to be set", ValidationError)
    if 0 < message.value < 1:
        warnings.warn(
            "Percentage values are 0-100, not fractions " f"({message.value} used)",
            ValidationWarning,
        )
    if message.value < 0 or message.value > 100:
        warnings.warn(
            f"Percentage value ({message.value}) is outside the expected range (0-100)",
            ValidationWarning,
        )
    ensure_float_nonnegative(message, "precision")


def validate_float_value(message: ord_schema.Message):
    ensure_float_nonnegative(message, "value")
    ensure_float_nonnegative(message, "precision")


def validate_data(message: reaction_pb2.Data):
    # TODO(kearnes): Validate/ping URLs?
    if not message.WhichOneof("kind"):
        warnings.warn("Data requires one of {value, bytes_value, url}", ValidationError)
    if message.bytes_value and not message.format:
        warnings.warn("Data format is required for bytes_data", ValidationError)


# pylint: enable=missing-function-docstring

_VALIDATOR_SWITCH = {
    dataset_pb2.Dataset: validate_dataset,
    dataset_pb2.DatasetExample: validate_dataset_example,
    reaction_pb2.Reaction: validate_reaction,
    # Basics
    reaction_pb2.ReactionIdentifier: validate_reaction_identifier,
    reaction_pb2.ReactionInput: validate_reaction_input,
    reaction_pb2.ReactionInput.AdditionDevice: validate_addition_device,
    reaction_pb2.ReactionInput.AdditionSpeed: validate_addition_speed,
    # Compounds
    reaction_pb2.Amount: validate_amount,
    reaction_pb2.UnmeasuredAmount: validate_unmeasured_amount,
    reaction_pb2.CrudeComponent: validate_crude_component,
    reaction_pb2.Compound: validate_compound,
    reaction_pb2.CompoundPreparation: validate_compound_preparation,
    reaction_pb2.CompoundIdentifier: validate_compound_identifier,
    reaction_pb2.Compound.Source: validate_source,
    # Setup
    reaction_pb2.Vessel: validate_vessel,
    reaction_pb2.VesselMaterial: validate_vessel_material,
    reaction_pb2.VesselAttachment: validate_vessel_attachment,
    reaction_pb2.VesselPreparation: validate_vessel_preparation,
    reaction_pb2.ReactionSetup: validate_reaction_setup,
    reaction_pb2.ReactionSetup.ReactionEnvironment: validate_reaction_environment,
    # Conditions
    reaction_pb2.ReactionConditions: validate_reaction_conditions,
    reaction_pb2.TemperatureConditions: validate_temperature_conditions,
    reaction_pb2.TemperatureConditions.TemperatureControl: validate_temperature_control,
    reaction_pb2.TemperatureConditions.Measurement: validate_temperature_measurement,
    reaction_pb2.PressureConditions: validate_pressure_conditions,
    reaction_pb2.PressureConditions.PressureControl: validate_pressure_control,
    reaction_pb2.PressureConditions.Atmosphere: validate_pressure_control,
    reaction_pb2.PressureConditions.Measurement: validate_pressure_measurement,
    reaction_pb2.StirringConditions: validate_stirring_conditions,
    reaction_pb2.StirringConditions.StirringRate: validate_stirring_rate,
    reaction_pb2.IlluminationConditions: validate_illumination_conditions,
    reaction_pb2.ElectrochemistryConditions: validate_electrochemistry_conditions,
    reaction_pb2.ElectrochemistryConditions.ElectrochemistryCell: validate_electrochemistry_cell,
    reaction_pb2.ElectrochemistryConditions.Measurement: validate_electrochemistry_measurement,
    reaction_pb2.FlowConditions: validate_flow_conditions,
    reaction_pb2.FlowConditions.Tubing: validate_tubing,
    # Annotations
    reaction_pb2.ReactionNotes: validate_reaction_notes,
    reaction_pb2.ReactionObservation: validate_reaction_observation,
    # Outcome
    reaction_pb2.ReactionWorkup: validate_reaction_workup,
    reaction_pb2.ReactionOutcome: validate_reaction_outcome,
    reaction_pb2.ProductCompound: validate_product_compound,
    reaction_pb2.ProductCompound.Texture: validate_texture,
    reaction_pb2.ProductMeasurement: validate_product_measurement,
    reaction_pb2.ProductMeasurement.Selectivity: validate_selectivity,
    reaction_pb2.ProductMeasurement.MassSpecMeasurementDetails: validate_mass_spec_measurement_type,
    reaction_pb2.DateTime: validate_date_time,
    reaction_pb2.Analysis: validate_analysis,
    # Metadata
    reaction_pb2.ReactionProvenance: validate_reaction_provenance,
    reaction_pb2.RecordEvent: validate_record_event,
    reaction_pb2.Person: validate_person,
    # Units
    reaction_pb2.Time: validate_time,
    reaction_pb2.Mass: validate_mass,
    reaction_pb2.Moles: validate_moles,
    reaction_pb2.Volume: validate_volume,
    reaction_pb2.Concentration: validate_concentration,
    reaction_pb2.Pressure: validate_pressure,
    reaction_pb2.Temperature: validate_temperature,
    reaction_pb2.Current: validate_current,
    reaction_pb2.Voltage: validate_voltage,
    reaction_pb2.Length: validate_length,
    reaction_pb2.Wavelength: validate_wavelength,
    reaction_pb2.FlowRate: validate_flow_rate,
    reaction_pb2.Percentage: validate_percentage,
    reaction_pb2.FloatValue: validate_float_value,
    reaction_pb2.Data: validate_data,
}
